# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-NRPC] Netlogon Remote Protocol

https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-nrpc/ff8f970f-3e37-40f7-bd4b-af7336e4792f
"""

import enum
import os
import struct
import time

from scapy.config import conf, crypto_validator
from scapy.fields import FlagValue, FlagsField
from scapy.layers.dcerpc import (
    find_dcerpc_interface,
    DCE_C_AUTHN_LEVEL,
    NL_AUTH_MESSAGE,
    NL_AUTH_SIGNATURE,
)
from scapy.layers.kerberos import KerberosSSP, _parse_upn
from scapy.layers.gssapi import (
    GSS_C_FLAGS,
    GSS_C_NO_CHANNEL_BINDINGS,
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    GSS_S_FAILURE,
    GSS_S_FLAGS,
    SSP,
)
from scapy.layers.ntlm import RC4, RC4K, RC4Init, MD4le

from scapy.layers.msrpce.rpcclient import (
    DCERPC_Client,
    DCERPC_Transport,
    STATUS_ERREF,
)
from scapy.layers.msrpce.raw.ms_nrpc import (
    NetrServerAuthenticate3_Request,
    NetrServerAuthenticate3_Response,
    NetrServerAuthenticateKerberos_Request,
    NetrServerAuthenticateKerberos_Response,
    NetrServerReqChallenge_Request,
    NetrServerReqChallenge_Response,
    NETLOGON_SECURE_CHANNEL_TYPE,
    PNETLOGON_AUTHENTICATOR,
    PNETLOGON_CREDENTIAL,
)


if conf.crypto_valid:
    from cryptography.hazmat.primitives import hashes, hmac
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
    from scapy.libs.rfc3961 import DES

    try:
        # cryptography > 47.0
        from cryptography.hazmat.decrepit.ciphers.modes import CFB8
    except ImportError:
        from cryptography.hazmat.primitives.ciphers.modes import CFB8
else:
    hashes = hmac = Cipher = algorithms = modes = DES = CFB8 = None


# Typing imports
from typing import (
    Optional,
)


# --- RFC

# [MS-NRPC] sect 3.1.4.2
_negotiateFlags = {
    # Not used. MUST be ignored on receipt.
    0x00000001: "A",
    # B: BDCs persistently try to update their database to the PDC's
    # version after they get a notification indicating that their
    # database is out-of-date.
    0x00000002: "BDCContinuousUpdate",
    # C: Supports RC4 encryption.
    0x00000004: "RC4",
    # Not used. MUST be ignored on receipt.
    0x00000008: "D",
    # E: Supports BDCs handling CHANGELOGs.
    0x00000010: "BDCChangelog",
    # F: Supports restarting of full synchronization between DCs.
    0x00000020: "RestartingDCSync",
    # G: Does not require ValidationLevel 2 fornongeneric passthrough.
    0x00000040: "NoValidationLevel2",
    # H: Supports the NetrDatabaseRedo (Opnum 17) functionality
    0x00000080: "DatabaseRedo",
    # I: Supports refusal of password changes.
    0x00000100: "RefusalPasswordChange",
    # J: Supports the NetrLogonSendToSam (Opnum 32) functionality.
    0x00000200: "SendToSam",
    # K: Supports generic pass-through authentication.
    0x00000400: "Generic-passthrough",
    # L: Supports concurrent RPC calls.
    0x00000800: "ConcurrentRPC",
    # M: Supports avoiding of user account database replication.
    0x00001000: "AvoidRepliAccountDB",
    # N: Supports avoiding of Security Authority database replication.
    0x00002000: "AvoidRepliAuthorityDB",
    # O: Supports strong keys.
    0x00004000: "StrongKeys",
    # P: Supports transitive trusts.
    0x00008000: "TransitiveTrust",
    # Not used. MUST be ignored on receipt.
    0x00010000: "Q",
    # R: Supports the NetrServerPasswordSet2 functionality.
    0x00020000: "ServerPasswordSet2",
    # S: Supports the NetrLogonGetDomainInfo functionality.
    0x00040000: "GetDomainInfo",
    # T: Supports cross-forest trusts.
    0x00080000: "CrossForestTrust",
    # U: The server ignores the NT4Emulator ADM element.
    0x00100000: "NoNT4Emul",
    # V: Supports RODC pass-through to different domains.
    0x00200000: "RODC-passthrough",
    # W: Supports Advanced Encryption Standard (AES) encryption and SHA2 hashing.
    0x01000000: "AES",
    # Not used. MUST be ignored on receipt.
    0x20000000: "X",
    # Y: Supports Secure RPC.
    0x40000000: "SecureRPC",
    # Supports Kerberos as the security support provider for secure channel setup.
    0x80000000: "Kerberos",
}
_negotiateFlags = FlagsField("", 0, -32, _negotiateFlags).names

# -- CRYPTO


# [MS-NRPC] sect 3.1.4.3.1
@crypto_validator
def ComputeSessionKeyAES(HashNt, ClientChallenge, ServerChallenge):
    M4SS = HashNt
    h = hmac.HMAC(M4SS, hashes.SHA256())
    h.update(ClientChallenge)
    h.update(ServerChallenge)
    return h.finalize()[:16]


# [MS-NRPC] sect 3.1.4.3.2
@crypto_validator
def ComputeSessionKeyStrongKey(HashNt, ClientChallenge, ServerChallenge):
    M4SS = HashNt
    digest = hashes.Hash(hashes.MD5())
    digest.update(b"\x00\x00\x00\x00")
    digest.update(ClientChallenge)
    digest.update(ServerChallenge)
    h = hmac.HMAC(M4SS, hashes.MD5())
    h.update(digest.finalize())
    return h.finalize()


# [MS-NRPC] sect 3.1.4.4.1
@crypto_validator
def ComputeNetlogonCredentialAES(Input, Sk):
    cipher = Cipher(algorithms.AES(Sk), mode=CFB8(b"\x00" * 16))
    encryptor = cipher.encryptor()
    return encryptor.update(Input)


# [MS-NRPC] sect 3.1.4.4.2
def InitLMKey(KeyIn):
    KeyOut = bytearray(b"\x00" * 8)
    KeyOut[0] = KeyIn[0] >> 0x01
    KeyOut[1] = ((KeyIn[0] & 0x01) << 6) | (KeyIn[1] >> 2)
    KeyOut[2] = ((KeyIn[1] & 0x03) << 5) | (KeyIn[2] >> 3)
    KeyOut[3] = ((KeyIn[2] & 0x07) << 4) | (KeyIn[3] >> 4)
    KeyOut[4] = ((KeyIn[3] & 0x0F) << 3) | (KeyIn[4] >> 5)
    KeyOut[5] = ((KeyIn[4] & 0x1F) << 2) | (KeyIn[5] >> 6)
    KeyOut[6] = ((KeyIn[5] & 0x3F) << 1) | (KeyIn[6] >> 7)
    KeyOut[7] = KeyIn[6] & 0x7F
    for i in range(8):
        KeyOut[i] = (KeyOut[i] << 1) & 0xFE
    return KeyOut


@crypto_validator
def ComputeNetlogonCredentialDES(Input, Sk):
    k3 = InitLMKey(Sk[0:7])
    k4 = InitLMKey(Sk[7:14])
    output1 = Cipher(DES(k3), modes.ECB()).encryptor().update(Input)
    return Cipher(DES(k4), modes.ECB()).encryptor().update(output1)


# [MS-NRPC] sect 3.1.4.5
def _credentialAddition(cred, i):
    return (
        struct.pack(
            "<I",
            (i + struct.unpack("<I", cred[:4])[0]) & 0xFFFFFFFF,
        )
        + cred[4:]
    )


# [MS-NRPC] sect 3.3.4.2.1


def ComputeCopySeqNumber(ClientSequenceNumber, client):
    low = struct.pack(">L", ClientSequenceNumber & 0xFFFFFFFF)
    high = struct.pack(
        ">L",
        ((ClientSequenceNumber >> 32) & 0xFFFFFFFF) | (0x80000000 if client else 0),
    )
    return low + high


@crypto_validator
def ComputeNetlogonChecksumAES(nl_auth_sig, message, SessionKey, Confounder=None):
    h = hmac.HMAC(SessionKey, hashes.SHA256())
    h.update(nl_auth_sig[:8])
    if Confounder:
        h.update(Confounder)
    h.update(message)
    return h.finalize()


@crypto_validator
def ComputeNetlogonChecksumMD5(nl_auth_sig, message, SessionKey, Confounder=None):
    digest = hashes.Hash(hashes.MD5())
    digest.update(b"\x00\x00\x00\x00")
    digest.update(nl_auth_sig[:8])
    if Confounder:
        digest.update(Confounder)
    digest.update(message)
    h = hmac.HMAC(SessionKey, hashes.MD5())
    h.update(digest.finalize())
    return h.finalize()


@crypto_validator
def ComputeNetlogonSealingKeyAES(SessionKey):
    return bytes(bytearray((x ^ 0xF0) for x in bytearray(SessionKey)))


@crypto_validator
def ComputeNetlogonSealingKeyRC4(SessionKey, CopySeqNumber):
    XorKey = bytes(bytearray((x ^ 0xF0) for x in bytearray(SessionKey)))
    h = hmac.HMAC(XorKey, hashes.MD5())
    h.update(b"\x00\x00\x00\x00")
    h = hmac.HMAC(h.finalize(), hashes.MD5())
    h.update(CopySeqNumber)
    return h.finalize()


@crypto_validator
def ComputeNetlogonSequenceNumberKeyMD5(SessionKey, Checksum):
    h = hmac.HMAC(SessionKey, hashes.MD5())
    h.update(b"\x00\x00\x00\x00")
    h = hmac.HMAC(h.finalize(), hashes.MD5())
    h.update(Checksum)
    return h.finalize()


# --- SSP


class NetlogonSSP(SSP):
    auth_type = 0x44  # Netlogon

    class STATE(SSP.STATE):
        INIT = 1
        CLI_SENT_NL = 2
        SRV_SENT_NL = 3

    class CONTEXT(SSP.CONTEXT):
        __slots__ = [
            "ClientSequenceNumber",
            "IsClient",
            "AES",
        ]

        def __init__(self, IsClient, req_flags=None, AES=True):
            self.state = NetlogonSSP.STATE.INIT
            self.IsClient = IsClient
            self.ClientSequenceNumber = 0
            self.AES = AES
            super(NetlogonSSP.CONTEXT, self).__init__(req_flags=req_flags)

    def __init__(self, SessionKey, computername, domainname, AES=True, **kwargs):
        self.SessionKey = SessionKey
        self.AES = AES
        self.computername = computername
        self.domainname = domainname
        super(NetlogonSSP, self).__init__(**kwargs)

    def GSS_Inquire_names_for_mech(self):
        raise NotImplementedError("Netlogon cannot be used with SPNEGO !")

    def _secure(self, Context, msgs, Seal):
        """
        Internal function used by GSS_WrapEx and GSS_GetMICEx

        [MS-NRPC] 3.3.4.2.1
        """
        # Concatenate the ToSign
        ToSign = b"".join(x.data for x in msgs if x.sign)

        Confounder = None
        if Seal:
            Confounder = os.urandom(8)

        if Context.AES:
            # 1. If AES is negotiated
            signature = NL_AUTH_SIGNATURE(
                SignatureAlgorithm=0x0013,
                SealAlgorithm=0x001A if Seal else 0xFFFF,
            )
        else:
            # 2. If AES is not negotiated
            signature = NL_AUTH_SIGNATURE(
                SignatureAlgorithm=0x0077,
                SealAlgorithm=0x007A if Seal else 0xFFFF,
            )
        # 3. Pad filled with 0xff (OK)
        # 4. Flags with 0x00 (OK)
        # 5. SequenceNumber
        SequenceNumber = ComputeCopySeqNumber(
            Context.ClientSequenceNumber, Context.IsClient
        )
        # 6. The ClientSequenceNumber MUST be incremented by 1
        Context.ClientSequenceNumber += 1
        # 7. Signature
        if Context.AES:
            signature.Checksum = ComputeNetlogonChecksumAES(
                bytes(signature), ToSign, self.SessionKey, Confounder
            )[:8]
        else:
            signature.Checksum = ComputeNetlogonChecksumMD5(
                bytes(signature), ToSign, self.SessionKey, Confounder
            )[:8]
        # 8. If the Confidentiality option is requested, the Confounder field and
        # the data MUST be encrypted
        if Seal:
            if Context.AES:
                EncryptionKey = ComputeNetlogonSealingKeyAES(self.SessionKey)
            else:
                EncryptionKey = ComputeNetlogonSealingKeyRC4(
                    self.SessionKey, SequenceNumber
                )
            # Encrypt Confounder and data
            if Context.AES:
                IV = SequenceNumber * 2
                encryptor = Cipher(
                    algorithms.AES(EncryptionKey), mode=CFB8(IV)
                ).encryptor()
                # Confounder
                signature.Confounder = encryptor.update(Confounder)
                # data
                for msg in msgs:
                    if msg.conf_req_flag:
                        msg.data = encryptor.update(msg.data)
            else:
                handle = RC4Init(EncryptionKey)
                # Confounder
                signature.Confounder = RC4(handle, Confounder)
                # DOC IS WRONG !
                # > The server MUST initialize RC4 only once, before encrypting
                # > the Confounder field.
                # But, this fails ! as Samba put it:
                # > For RC4, Windows resets the cipherstate after encrypting
                # > the confounder, thus defeating the purpose of the confounder
                handle = RC4Init(EncryptionKey)
                # data
                for msg in msgs:
                    if msg.conf_req_flag:
                        msg.data = RC4(handle, msg.data)
        # 9. The SequenceNumber MUST be encrypted.
        if Context.AES:
            EncryptionKey = self.SessionKey
            IV = signature.Checksum * 2
            cipher = Cipher(algorithms.AES(EncryptionKey), mode=CFB8(IV))
            encryptor = cipher.encryptor()
            signature.SequenceNumber = encryptor.update(SequenceNumber)
        else:
            EncryptionKey = ComputeNetlogonSequenceNumberKeyMD5(
                self.SessionKey, signature.Checksum
            )
            signature.SequenceNumber = RC4K(EncryptionKey, SequenceNumber)

        return (
            msgs,
            signature,
        )

    def _unsecure(self, Context, msgs, signature, Seal):
        """
        Internal function used by GSS_UnwrapEx and GSS_VerifyMICEx

        [MS-NRPC] 3.3.4.2.2
        """
        assert isinstance(signature, NL_AUTH_SIGNATURE)

        # 1. The SignatureAlgorithm bytes MUST be verified
        if (Context.AES and signature.SignatureAlgorithm != 0x0013) or (
            not Context.AES and signature.SignatureAlgorithm != 0x0077
        ):
            raise ValueError("Invalid SignatureAlgorithm !")

        # 5. The SequenceNumber MUST be decrypted.
        if Context.AES:
            EncryptionKey = self.SessionKey
            IV = signature.Checksum * 2
            cipher = Cipher(algorithms.AES(EncryptionKey), mode=CFB8(IV))
            decryptor = cipher.decryptor()
            SequenceNumber = decryptor.update(signature.SequenceNumber)
        else:
            EncryptionKey = ComputeNetlogonSequenceNumberKeyMD5(
                self.SessionKey, signature.Checksum
            )
            SequenceNumber = RC4K(EncryptionKey, signature.SequenceNumber)
        # 6. A local copy of SequenceNumber MUST be computed
        CopySeqNumber = ComputeCopySeqNumber(
            Context.ClientSequenceNumber, not Context.IsClient
        )
        # 7. The SequenceNumber MUST be compared to CopySeqNumber
        if SequenceNumber != CopySeqNumber:
            raise ValueError("ERROR: SequenceNumber don't match")
        # 8. ClientSequenceNumber MUST be incremented.
        Context.ClientSequenceNumber += 1
        # 9. If the Confidentiality option is requested, the Confounder and the
        # data MUST be decrypted.
        Confounder = None
        if Seal:
            if Context.AES:
                EncryptionKey = ComputeNetlogonSealingKeyAES(self.SessionKey)
            else:
                EncryptionKey = ComputeNetlogonSealingKeyRC4(
                    self.SessionKey, SequenceNumber
                )
            # Decrypt Confounder and data
            if Context.AES:
                IV = SequenceNumber * 2
                decryptor = Cipher(
                    algorithms.AES(EncryptionKey), mode=CFB8(IV)
                ).decryptor()
                # Confounder
                Confounder = decryptor.update(signature.Confounder)
                # data
                for msg in msgs:
                    if msg.conf_req_flag:
                        msg.data = decryptor.update(msg.data)
            else:
                # Confounder
                EncryptionKey = ComputeNetlogonSealingKeyRC4(
                    self.SessionKey, SequenceNumber
                )
                Confounder = RC4K(EncryptionKey, signature.Confounder)
                # data
                handle = RC4Init(EncryptionKey)
                for msg in msgs:
                    if msg.conf_req_flag:
                        msg.data = RC4(handle, msg.data)

        # Concatenate the ToSign
        ToSign = b"".join(x.data for x in msgs if x.sign)

        # 10/11. Signature
        if Context.AES:
            Checksum = ComputeNetlogonChecksumAES(
                bytes(signature), ToSign, self.SessionKey, Confounder
            )[:8]
        else:
            Checksum = ComputeNetlogonChecksumMD5(
                bytes(signature), ToSign, self.SessionKey, Confounder
            )[:8]
        if signature.Checksum != Checksum:
            raise ValueError("ERROR: Checksum don't match")
        return msgs

    def GSS_WrapEx(self, Context, msgs, qop_req=0):
        return self._secure(Context, msgs, True)

    def GSS_GetMICEx(self, Context, msgs, qop_req=0):
        return self._secure(Context, msgs, False)[1]

    def GSS_UnwrapEx(self, Context, msgs, signature):
        return self._unsecure(Context, msgs, signature, True)

    def GSS_VerifyMICEx(self, Context, msgs, signature):
        self._unsecure(Context, msgs, signature, False)

    def GSS_Init_sec_context(
        self,
        Context: CONTEXT,
        input_token=None,
        target_name: Optional[str] = None,
        req_flags: Optional[GSS_C_FLAGS] = None,
        chan_bindings: bytes = GSS_C_NO_CHANNEL_BINDINGS,
    ):
        if Context is None:
            Context = self.CONTEXT(True, req_flags=req_flags, AES=self.AES)

        if Context.state == self.STATE.INIT:
            Context.state = self.STATE.CLI_SENT_NL
            return (
                Context,
                NL_AUTH_MESSAGE(
                    MessageType=0,
                    Flags=3,
                    NetbiosDomainName=self.domainname,
                    NetbiosComputerName=self.computername,
                ),
                GSS_S_CONTINUE_NEEDED,
            )
        else:
            return Context, None, GSS_S_COMPLETE

    def GSS_Accept_sec_context(
        self,
        Context: CONTEXT,
        input_token=None,
        req_flags: Optional[GSS_S_FLAGS] = GSS_S_FLAGS.GSS_S_ALLOW_MISSING_BINDINGS,
        chan_bindings: bytes = GSS_C_NO_CHANNEL_BINDINGS,
    ):
        if Context is None:
            Context = self.CONTEXT(False, req_flags=req_flags, AES=self.AES)

        if Context.state == self.STATE.INIT:
            Context.state = self.STATE.SRV_SENT_NL
            return (
                Context,
                NL_AUTH_MESSAGE(
                    MessageType=1,
                    Flags=0,
                ),
                GSS_S_COMPLETE,
            )
        else:
            # Invalid state
            return Context, None, GSS_S_FAILURE

    def MaximumSignatureLength(self, Context: CONTEXT):
        """
        Returns the Maximum Signature length.

        This will be used in auth_len in DceRpc5, and is necessary for
        PFC_SUPPORT_HEADER_SIGN to work properly.
        """
        # len(NL_AUTH_SIGNATURE())
        if Context.flags & GSS_C_FLAGS.GSS_C_CONF_FLAG:
            if Context.AES:
                return 56
            else:
                return 32
        else:
            if Context.AES:
                return 48
            else:
                return 24


# --- Utils


class NETLOGON_SECURE_CHANNEL_METHOD(enum.Enum):
    NetrServerAuthenticate3 = 1
    NetrServerAuthenticateKerberos = 2


class NetlogonClient(DCERPC_Client):
    """
    A subclass of DCERPC_Client that supports establishing a Netlogon secure channel
    using the Netlogon SSP, and handling Netlogon authenticators.

    This class therefore only supports the 'logon' rpc.

    :param auth_level: one of DCE_C_AUTHN_LEVEL

    :param verb: verbosity control.
    :param supportAES: advertise AES support in the Netlogon session.

    Example::

        >>> cli = NetlogonClient()
        >>> cli.connect_and_bind("192.168.0.100")
        >>> cli.establish_secure_channel(
        ...     UPN="WIN10@DOMAIN",
        ...     HASHNT=bytes.fromhex("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
        ... )
    """

    def __init__(
        self,
        # Default to PRIVACY: see KB5021130
        auth_level=DCE_C_AUTHN_LEVEL.PKT_PRIVACY,
        verb=True,
        supportAES=True,
        **kwargs,
    ):
        self.interface = find_dcerpc_interface("logon")
        self.SessionKey = None
        self.ClientStoredCredential = None
        self.supportAES = supportAES
        super(NetlogonClient, self).__init__(
            DCERPC_Transport.NCACN_IP_TCP,
            auth_level=auth_level,
            verb=verb,
            **kwargs,
        )

    def connect(self, host, **kwargs):
        """
        This calls DCERPC_Client's connect_and_bind to bind the 'logon' interface.
        """
        super(NetlogonClient, self).connect(
            host=host,
            interface=self.interface,
            **kwargs,
        )

    def create_authenticator(self):
        """
        Create a NETLOGON_AUTHENTICATOR
        """
        # [MS-NRPC] sect 3.1.4.5
        ts = int(time.time())
        self.ClientStoredCredential = _credentialAddition(
            self.ClientStoredCredential, ts
        )
        return PNETLOGON_AUTHENTICATOR(
            Credential=PNETLOGON_CREDENTIAL(
                data=(
                    ComputeNetlogonCredentialAES(
                        self.ClientStoredCredential,
                        self.SessionKey,
                    )
                    if self.supportAES
                    else ComputeNetlogonCredentialDES(
                        self.ClientStoredCredential,
                        self.SessionKey,
                    )
                ),
            ),
            Timestamp=ts,
        )

    def validate_authenticator(self, auth):
        """
        Validate a NETLOGON_AUTHENTICATOR

        :param auth: the NETLOGON_AUTHENTICATOR object
        """
        # [MS-NRPC] sect 3.1.4.5
        self.ClientStoredCredential = _credentialAddition(
            self.ClientStoredCredential, 1
        )
        if self.supportAES:
            tempcred = ComputeNetlogonCredentialAES(
                self.ClientStoredCredential, self.SessionKey
            )
        else:
            tempcred = ComputeNetlogonCredentialDES(
                self.ClientStoredCredential, self.SessionKey
            )
        if tempcred != auth.Credential.data:
            raise ValueError("Server netlogon authenticator is wrong !")

    def establish_secure_channel(
        self,
        UPN: str,
        DC_FQDN: str,
        HASHNT: Optional[bytes] = None,
        PASSWORD: Optional[str] = None,
        KEY=None,
        ssp: Optional[KerberosSSP] = None,
        mode=NETLOGON_SECURE_CHANNEL_METHOD.NetrServerAuthenticate3,
        secureChannelType=NETLOGON_SECURE_CHANNEL_TYPE.WorkstationSecureChannel,
    ):
        """
        Function to establish the Netlogon Secure Channel.

        This uses NetrServerAuthenticate3 or NetrServerAuthenticateKerberos to
        negotiate the session key, then creates a NetlogonSSP that uses that session
        key and alters the DCE/RPC session to use it.

        :param mode: one of NETLOGON_SECURE_CHANNEL_METHOD. This defines which method
                     to use to establish the secure channel.
        :param UPN: the UPN of the computer account name that is used to establish
                    the secure channel. (e.g. WIN10$@domain.local)
        :param DC_FQDN: the FQDN name of the DC.

        The function then requires one of the following:

        :param HASHNT: the HashNT of the computer account (in Authenticate3 mode).
        :param KEY: a Kerberos key to use (in Kerberos mode)
        :param PASSWORD: the password of the computer account (any mode).
        :param ssp: a KerberosSSP to use (in Kerberos mode)
        """
        computername, domainname = _parse_upn(UPN)
        # We need to normalize here, since the functions require both the accountname
        # and the normal (no dollar) computer name.
        if computername.endswith("$"):
            computername = computername[:-1]

        if mode == NETLOGON_SECURE_CHANNEL_METHOD.NetrServerAuthenticate3:
            if ssp or KEY:
                raise ValueError("Cannot use 'ssp' on 'KEY' in Authenticate3 mode !")
            if not HASHNT:
                if PASSWORD:
                    HASHNT = MD4le(PASSWORD)
                else:
                    raise ValueError("Missing either 'PASSWORD' or 'HASHNT' !")
            if "." in domainname:
                raise ValueError(
                    "The UPN in Authenticate3 must have a NETBIOS domain name !"
                )
        else:
            if HASHNT:
                raise ValueError("Cannot use 'HASHNT' in Kerberos mode !")

        # Calc NegotiateFlags
        NegotiateFlags = FlagValue(
            0x602FFFFF,  # sensible default (Windows)
            names=_negotiateFlags,
        )
        if self.supportAES:
            NegotiateFlags += "AES"

        # We are either using NetrServerAuthenticate3 or NetrServerAuthenticateKerberos
        if mode == NETLOGON_SECURE_CHANNEL_METHOD.NetrServerAuthenticate3:
            # We use the legacy NetrServerAuthenticate3 function (NetlogonSSP)

            # Make sure the interface is bound
            if not self.bind_or_alter(self.interface):
                raise ValueError("Bind failed !")

            # Flow documented in 3.1.4 Session-Key Negotiation
            # and sect 3.4.5.2 for specific calls
            clientChall = os.urandom(8)

            # Perform NetrServerReqChallenge request
            netr_server_req_chall_response = self.sr1_req(
                NetrServerReqChallenge_Request(
                    PrimaryName=None,
                    ComputerName=computername,
                    ClientChallenge=PNETLOGON_CREDENTIAL(
                        data=clientChall,
                    ),
                    ndr64=self.ndr64,
                    ndrendian=self.ndrendian,
                )
            )
            if (
                NetrServerReqChallenge_Response not in netr_server_req_chall_response
                or netr_server_req_chall_response.status != 0
            ):
                print(
                    conf.color_theme.fail(
                        "! %s"
                        % STATUS_ERREF.get(
                            netr_server_req_chall_response.status, "Failure"
                        )
                    )
                )
                netr_server_req_chall_response.show()
                raise ValueError("NetrServerReqChallenge failed !")

            # Build the session key
            serverChall = netr_server_req_chall_response.ServerChallenge.data
            if self.supportAES:
                SessionKey = ComputeSessionKeyAES(HASHNT, clientChall, serverChall)
                self.ClientStoredCredential = ComputeNetlogonCredentialAES(
                    clientChall, SessionKey
                )
            else:
                SessionKey = ComputeSessionKeyStrongKey(
                    HASHNT, clientChall, serverChall
                )
                self.ClientStoredCredential = ComputeNetlogonCredentialDES(
                    clientChall, SessionKey
                )

            # Perform Authenticate3 request
            netr_server_auth3_response = self.sr1_req(
                NetrServerAuthenticate3_Request(
                    PrimaryName="\\\\" + DC_FQDN,
                    AccountName=computername + "$",
                    SecureChannelType=secureChannelType,
                    ComputerName=computername,
                    ClientCredential=PNETLOGON_CREDENTIAL(
                        data=self.ClientStoredCredential,
                    ),
                    NegotiateFlags=int(NegotiateFlags),
                    ndr64=self.ndr64,
                    ndrendian=self.ndrendian,
                )
            )
            if netr_server_auth3_response.status != 0:
                # An error occurred.
                NegotiatedFlags = None
                if NetrServerAuthenticate3_Response in netr_server_auth3_response:
                    NegotiatedFlags = FlagValue(
                        netr_server_auth3_response.NegotiateFlags,
                        names=_negotiateFlags,
                    )
                    if NegotiateFlags != NegotiatedFlags:
                        print(
                            conf.color_theme.fail(
                                "! Unsupported server flags: %s"
                                % (NegotiatedFlags ^ NegotiateFlags)
                            )
                        )
                raise ValueError("NetrServerAuthenticate3 failed !")

            # Check Server Credential
            if self.supportAES:
                if (
                    netr_server_auth3_response.ServerCredential.data
                    != ComputeNetlogonCredentialAES(serverChall, SessionKey)
                ):
                    print(conf.color_theme.fail("! Invalid ServerCredential."))
                    raise ValueError
            else:
                if (
                    netr_server_auth3_response.ServerCredential.data
                    != ComputeNetlogonCredentialDES(serverChall, SessionKey)
                ):
                    print(conf.color_theme.fail("! Invalid ServerCredential."))
                    raise ValueError

            # SessionKey negotiated !
            self.SessionKey = SessionKey

            # Create the NetlogonSSP and assign it to the local client
            self.ssp = self.sock.session.ssp = NetlogonSSP(
                SessionKey=self.SessionKey,
                AES=self.supportAES,
                domainname=domainname,
                computername=computername,
            )

            # Finally alter context (to use the SSP)
            if not self.alter_context(self.interface):
                raise ValueError("Bind failed !")

        elif mode == NETLOGON_SECURE_CHANNEL_METHOD.NetrServerAuthenticateKerberos:
            # We use the brand new NetrServerAuthenticateKerberos function
            NegotiateFlags += "Kerberos"

            # Set KerberosSSP and alter context
            if ssp:
                self.ssp = self.sock.session.ssp = ssp
            else:
                self.ssp = self.sock.session.ssp = KerberosSSP(
                    UPN=UPN,
                    SPN="netlogon/" + DC_FQDN,
                    PASSWORD=PASSWORD,
                    KEY=KEY,
                )
            if not self.bind_or_alter(self.interface):
                raise ValueError("Bind failed !")

            # Send AuthenticateKerberos request
            netr_server_authkerb_response = self.sr1_req(
                NetrServerAuthenticateKerberos_Request(
                    PrimaryName="\\\\" + DC_FQDN,
                    AccountName=computername + "$",
                    AccountType=secureChannelType,
                    ComputerName=computername,
                    NegotiateFlags=int(NegotiateFlags),
                    ndr64=self.ndr64,
                    ndrendian=self.ndrendian,
                )
            )
            if (
                NetrServerAuthenticateKerberos_Response
                not in netr_server_authkerb_response
                or netr_server_authkerb_response.status != 0
            ):
                # An error occurred
                netr_server_authkerb_response.show()
                raise ValueError("NetrServerAuthenticateKerberos failed !")

            # The NRPC session key is in this case the kerberos one
            self.SessionKey = self.sspcontext.SessionKey



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = AUTOSAR Secure On-Board Communication
# scapy.contrib.status = library

"""
SecOC
"""
from scapy.config import conf
from scapy.error import log_loading

if conf.crypto_valid:
    from cryptography.hazmat.primitives import cmac
    from cryptography.hazmat.primitives.ciphers import algorithms
else:
    log_loading.info("Can't import python-cryptography v2.0+. "
                     "Disabled SecOC calculate_cmac.")

from scapy.config import conf
from scapy.fields import PacketLenField
from scapy.packet import Packet, Raw

# Typing imports
from typing import (
    Callable,
    Dict,
    Optional,
    Set,
    Type,
)


class SecOCMixin:

    pdu_payload_cls_by_identifier: Dict[int, Type[Packet]] = dict()
    secoc_protected_pdus_by_identifier: Set[int] = set()

    def secoc_authenticate(self) -> None:
        raise NotImplementedError

    def secoc_verify(self) -> bool:
        raise NotImplementedError

    def get_secoc_payload(self) -> bytes:
        """Override this method for customization
        """
        raise NotImplementedError

    def get_secoc_key(self) -> bytes:
        """Override this method for customization
        """
        return b"\x00" * 16

    def get_secoc_freshness_value(self) -> bytes:
        """Override this method for customization
        """
        return b"\x00" * 4

    def get_message_authentication_code(self):
        payload = self.get_secoc_payload()
        key = self.get_secoc_key()
        freshness_value = self.get_secoc_freshness_value()
        return self.calculate_cmac(key, payload, freshness_value)

    @staticmethod
    def calculate_cmac(key: bytes, payload: bytes, freshness_value: bytes) -> bytes:
        c = cmac.CMAC(algorithms.AES128(key))
        c.update(payload + freshness_value)
        return c.finalize()

    @classmethod
    def register_secoc_protected_pdu(cls,
                                     pdu_id: int,
                                     pdu_payload_cls: Type[Packet] = Raw
                                     ) -> None:
        cls.secoc_protected_pdus_by_identifier.add(pdu_id)
        cls.pdu_payload_cls_by_identifier[pdu_id] = pdu_payload_cls

    @classmethod
    def unregister_secoc_protected_pdu(cls, pdu_id: int) -> None:
        cls.secoc_protected_pdus_by_identifier.remove(pdu_id)
        del cls.pdu_payload_cls_by_identifier[pdu_id]


class PduPayloadField(PacketLenField):
    __slots__ = ["guess_pkt_cls"]

    def __init__(self,
                 name,  # type: str
                 default,  # type: Packet
                 guess_pkt_cls,  # type: Callable[[Packet, bytes], Packet]  # noqa: E501
                 length_from=None  # type: Optional[Callable[[Packet], int]]  # noqa: E501
                 ):
        # type: (...) -> None
        super(PacketLenField, self).__init__(name, default, Raw)
        self.length_from = length_from or (lambda x: 0)
        self.guess_pkt_cls = guess_pkt_cls

    def m2i(self, pkt, m):  # type: ignore
        # type: (Optional[Packet], bytes) -> Packet
        return self.guess_pkt_cls(pkt, m)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# Copyright (C) Ryan Speers <ryan@rmspeers.com> 2011-2012
# Copyright (C) Roger Meyer <roger.meyer@csus.edu>: 2012-03-10 Added frames
# Copyright (C) Gabriel Potter <gabriel[]potter[]fr>: 2018
# Copyright (C) Dimitrios-Georgios Akestoridis <akestoridis@cmu.edu>

"""
Wireless MAC according to IEEE 802.15.4.
"""

import struct

from scapy.compat import orb, chb
from scapy.error import warning
from scapy.config import conf

from scapy.data import DLT_IEEE802_15_4_WITHFCS, DLT_IEEE802_15_4_NOFCS
from scapy.packet import Packet, bind_layers
from scapy.fields import (
    BitEnumField,
    BitField,
    ByteEnumField,
    ByteField,
    ConditionalField,
    Emph,
    FCSField,
    Field,
    FieldListField,
    LELongField,
    MultipleTypeField,
    PacketField,
    StrFixedLenField,
    XByteField,
    XLEIntField,
    XLEShortField,
)

# Fields #


class dot15d4AddressField(Field):
    __slots__ = ["adjust", "length_of"]

    def __init__(self, name, default, length_of=None, fmt="<H", adjust=None):
        Field.__init__(self, name, default, fmt)
        self.length_of = length_of
        if adjust is not None:
            self.adjust = adjust
        else:
            self.adjust = lambda pkt, x: self.lengthFromAddrMode(pkt, x)

    def i2repr(self, pkt, x):
        """Convert internal value to a nice representation"""
        if len(hex(self.i2m(pkt, x))) < 7:  # short address
            return hex(self.i2m(pkt, x))
        else:  # long address
            x = "%016x" % self.i2m(pkt, x)
            return ":".join(["%s%s" % (x[i], x[i + 1]) for i in range(0, len(x), 2)])  # noqa: E501

    def addfield(self, pkt, s, val):
        """Add an internal value to a string"""
        if self.adjust(pkt, self.length_of) == 2:
            return s + struct.pack(self.fmt[0] + "H", val)
        elif self.adjust(pkt, self.length_of) == 8:
            return s + struct.pack(self.fmt[0] + "Q", val)
        else:
            return s

    def getfield(self, pkt, s):
        if self.adjust(pkt, self.length_of) == 2:
            return s[2:], self.m2i(pkt, struct.unpack(self.fmt[0] + "H", s[:2])[0])  # noqa: E501
        elif self.adjust(pkt, self.length_of) == 8:
            return s[8:], self.m2i(pkt, struct.unpack(self.fmt[0] + "Q", s[:8])[0])  # noqa: E501
        else:
            raise Exception('impossible case')

    def lengthFromAddrMode(self, pkt, x):
        addrmode = 0
        pkttop = pkt.underlayer
        if pkttop is None:
            warning("No underlayer to guess address mode")
            return 0
        while True:
            try:
                addrmode = pkttop.getfieldval(x)
                break
            except Exception:
                if pkttop.underlayer is None:
                    break
                pkttop = pkttop.underlayer
        # print("Underlayer field value of", x, "is", addrmode)
        if addrmode == 2:
            return 2
        elif addrmode == 3:
            return 8
        return 0


# Layers #

class Dot15d4(Packet):
    name = "802.15.4"
    fields_desc = [
        BitField("fcf_reserved_1", 0, 1),  # fcf p1 b1
        BitEnumField("fcf_panidcompress", 0, 1, [False, True]),
        BitEnumField("fcf_ackreq", 0, 1, [False, True]),
        BitEnumField("fcf_pending", 0, 1, [False, True]),
        BitEnumField("fcf_security", 0, 1, [False, True]),  # fcf p1 b2
        Emph(BitEnumField("fcf_frametype", 0, 3, {0: "Beacon", 1: "Data", 2: "Ack", 3: "Command"})),  # noqa: E501
        BitEnumField("fcf_srcaddrmode", 0, 2, {0: "None", 1: "Reserved", 2: "Short", 3: "Long"}),  # fcf p2 b1  # noqa: E501
        BitField("fcf_framever", 0, 2),  # 00 compatibility with 2003 version; 01 compatible with 2006 version  # noqa: E501
        BitEnumField("fcf_destaddrmode", 2, 2, {0: "None", 1: "Reserved", 2: "Short", 3: "Long"}),  # fcf p2 b2  # noqa: E501
        BitField("fcf_reserved_2", 0, 2),
        Emph(ByteField("seqnum", 1))  # sequence number
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 %Dot15d4.fcf_frametype% ackreq(%Dot15d4.fcf_ackreq%) ( %Dot15d4.fcf_destaddrmode% -> %Dot15d4.fcf_srcaddrmode% ) Seq#%Dot15d4.seqnum%")  # noqa: E501

    def guess_payload_class(self, payload):
        if self.fcf_frametype == 0x00:
            return Dot15d4Beacon
        elif self.fcf_frametype == 0x01:
            return Dot15d4Data
        elif self.fcf_frametype == 0x02:
            return Dot15d4Ack
        elif self.fcf_frametype == 0x03:
            return Dot15d4Cmd
        else:
            return Packet.guess_payload_class(self, payload)

    def answers(self, other):
        if isinstance(other, Dot15d4):
            if self.fcf_frametype == 2:  # ack
                if self.seqnum != other.seqnum:  # check for seqnum matching
                    return 0
                elif other.fcf_ackreq == 1:  # check that an ack was indeed requested  # noqa: E501
                    return 1
        return 0

    def post_build(self, p, pay):
        # This just forces destaddrmode to None for Ack frames.
        if self.fcf_frametype == 2 and self.fcf_destaddrmode != 0:
            self.fcf_destaddrmode = 0
            return p[:1] + \
                chb((self.fcf_srcaddrmode << 6) + (self.fcf_framever << 4)) \
                + p[2:] + pay
        else:
            return p + pay


class Dot15d4FCS(Dot15d4):
    '''
    This class is a drop-in replacement for the Dot15d4 class above, except
    it expects a FCS/checksum in the input, and produces one in the output.
    This provides the user flexibility, as many 802.15.4 interfaces will have an AUTO_CRC setting  # noqa: E501
    that will validate the FCS/CRC in firmware, and add it automatically when transmitting.  # noqa: E501
    '''
    name = "802.15.4 - FCS"
    match_subclass = True
    fields_desc = Dot15d4.fields_desc + [FCSField("fcs", None, fmt="<H")]

    def compute_fcs(self, data):
        # Do a CRC-CCITT Kermit 16bit on the data given
        # Returns a CRC that is the FCS for the frame
        #  Implemented using pseudocode from: June 1986, Kermit Protocol Manual
        #  See also:
        #   http://regregex.bbcmicro.net/crc-catalogue.htm#crc.cat.kermit
        crc = 0
        for i in range(0, len(data)):
            c = orb(data[i])
            q = (crc ^ c) & 15  # Do low-order 4 bits
            crc = (crc // 16) ^ (q * 4225)
            q = (crc ^ (c // 16)) & 15  # And high 4 bits
            crc = (crc // 16) ^ (q * 4225)
        return struct.pack('<H', crc)  # return as bytes in little endian order

    def post_build(self, p, pay):
        # construct the packet with the FCS at the end
        p = Dot15d4.post_build(self, p, pay)
        if self.fcs is None:
            p = p[:-2]
            p = p + self.compute_fcs(p)
        return p


class Dot15d4Ack(Packet):
    name = "802.15.4 Ack"
    fields_desc = []


class Dot15d4AuxSecurityHeader(Packet):
    name = "802.15.4 Auxiliary Security Header"
    fields_desc = [
        BitField("sec_sc_reserved", 0, 3),
        # Key Identifier Mode
        # 0: Key is determined implicitly from the originator and recipient(s) of the frame  # noqa: E501
        # 1: Key is determined explicitly from the the 1-octet Key Index subfield of the Key Identifier field  # noqa: E501
        # 2: Key is determined explicitly from the 4-octet Key Source and the 1-octet Key Index  # noqa: E501
        # 3: Key is determined explicitly from the 8-octet Key Source and the 1-octet Key Index  # noqa: E501
        BitEnumField("sec_sc_keyidmode", 0, 2, {
            0: "Implicit", 1: "1oKeyIndex", 2: "4o-KeySource-1oKeyIndex", 3: "8o-KeySource-1oKeyIndex"}  # noqa: E501
        ),
        BitEnumField("sec_sc_seclevel", 0, 3, {0: "None", 1: "MIC-32", 2: "MIC-64", 3: "MIC-128", 4: "ENC", 5: "ENC-MIC-32", 6: "ENC-MIC-64", 7: "ENC-MIC-128"}),  # noqa: E501
        XLEIntField("sec_framecounter", 0x00000000),  # 4 octets
        # Key Identifier (variable length): identifies the key that is used for cryptographic protection  # noqa: E501
        # Key Source : length of sec_keyid_keysource varies btwn 0, 4, and 8 bytes depending on sec_sc_keyidmode  # noqa: E501
        MultipleTypeField([
            # 4 octets when sec_sc_keyidmode == 2
            (XLEIntField("sec_keyid_keysource", 0x00000000),
                lambda pkt: pkt.getfieldval("sec_sc_keyidmode") == 2),
            # 8 octets when sec_sc_keyidmode == 3
            (LELongField("sec_keyid_keysource", 0x0000000000000000),
                lambda pkt: pkt.getfieldval("sec_sc_keyidmode") == 3),
        ], StrFixedLenField("sec_keyid_keysource", "", length=0)),
        # Key Index (1 octet): allows unique identification of different keys with the same originator  # noqa: E501
        ConditionalField(XByteField("sec_keyid_keyindex", 0xFF),
                         lambda pkt: pkt.getfieldval("sec_sc_keyidmode") != 0),
    ]


class Dot15d4Data(Packet):
    name = "802.15.4 Data"
    fields_desc = [
        XLEShortField("dest_panid", 0xFFFF),
        dot15d4AddressField("dest_addr", 0xFFFF, length_of="fcf_destaddrmode"),
        ConditionalField(XLEShortField("src_panid", 0x0),
                         lambda pkt:util_srcpanid_present(pkt)),
        ConditionalField(dot15d4AddressField("src_addr", None, length_of="fcf_srcaddrmode"),  # noqa: E501
                         lambda pkt:pkt.underlayer.getfieldval("fcf_srcaddrmode") != 0),  # noqa: E501
        # Security field present if fcf_security == True
        ConditionalField(PacketField("aux_sec_header", Dot15d4AuxSecurityHeader(), Dot15d4AuxSecurityHeader),  # noqa: E501
                         lambda pkt:pkt.underlayer.getfieldval("fcf_security") is True),  # noqa: E501
    ]

    def guess_payload_class(self, payload):
        # TODO: See how it's done in wireshark:
        # https://github.com/wireshark/wireshark/blob/93c60b3b7c801dddd11d8c7f2a0ea4b7d02d700a/epan/dissectors/packet-ieee802154.c#L2061  # noqa: E501
        # it's too magic to me
        from scapy.layers.sixlowpan import SixLoWPAN
        from scapy.layers.zigbee import ZigbeeNWK
        if conf.dot15d4_protocol == "sixlowpan":
            return SixLoWPAN
        elif conf.dot15d4_protocol == "zigbee":
            return ZigbeeNWK
        else:
            if conf.dot15d4_protocol is None:
                _msg = "Please set conf.dot15d4_protocol to select a " + \
                       "802.15.4 protocol. Values must be in the list: "
            else:
                _msg = "Unknown conf.dot15d4_protocol value: must be in "
            warning(_msg +
                    "['sixlowpan', 'zigbee']" +
                    " Defaulting to SixLoWPAN")
            return SixLoWPAN

    def mysummary(self):
        return self.sprintf("802.15.4 Data ( %Dot15d4Data.src_panid%:%Dot15d4Data.src_addr% -> %Dot15d4Data.dest_panid%:%Dot15d4Data.dest_addr% )")  # noqa: E501


class Dot15d4Beacon(Packet):
    name = "802.15.4 Beacon"
    fields_desc = [
        XLEShortField("src_panid", 0x0),
        dot15d4AddressField("src_addr", None, length_of="fcf_srcaddrmode"),
        # Security field present if fcf_security == True
        ConditionalField(PacketField("aux_sec_header", Dot15d4AuxSecurityHeader(), Dot15d4AuxSecurityHeader),  # noqa: E501
                         lambda pkt:pkt.underlayer.getfieldval("fcf_security") is True),  # noqa: E501

        # Superframe spec field:
        BitField("sf_sforder", 15, 4),  # not used by ZigBee
        BitField("sf_beaconorder", 15, 4),  # not used by ZigBee
        BitEnumField("sf_assocpermit", 0, 1, [False, True]),
        BitEnumField("sf_pancoord", 0, 1, [False, True]),
        BitField("sf_reserved", 0, 1),  # not used by ZigBee
        BitEnumField("sf_battlifeextend", 0, 1, [False, True]),  # not used by ZigBee  # noqa: E501
        BitField("sf_finalcapslot", 15, 4),  # not used by ZigBee

        # GTS Fields
        #  GTS Specification (1 byte)
        BitEnumField("gts_spec_permit", 1, 1, [False, True]),  # GTS spec bit 7, true=1 iff PAN cord is accepting GTS requests  # noqa: E501
        BitField("gts_spec_reserved", 0, 4),  # GTS spec bits 3-6
        BitField("gts_spec_desccount", 0, 3),  # GTS spec bits 0-2
        #  GTS Directions (0 or 1 byte)
        ConditionalField(BitField("gts_dir_reserved", 0, 1), lambda pkt:pkt.getfieldval("gts_spec_desccount") != 0),  # noqa: E501
        ConditionalField(BitField("gts_dir_mask", 0, 7), lambda pkt:pkt.getfieldval("gts_spec_desccount") != 0),  # noqa: E501
        #  GTS List (variable size)
        # TODO add a Packet/FieldListField tied to 3bytes per count in gts_spec_desccount  # noqa: E501

        # Pending Address Fields:
        #  Pending Address Specification (1 byte)
        BitField("pa_reserved_1", 0, 1),
        BitField("pa_num_long", 0, 3),  # number of long addresses pending
        BitField("pa_reserved_2", 0, 1),
        BitField("pa_num_short", 0, 3),  # number of short addresses pending
        #  Address List (var length)
        FieldListField("pa_short_addresses", [],
                       XLEShortField("", 0x0000),
                       count_from=lambda pkt: pkt.pa_num_short),
        FieldListField("pa_long_addresses", [],
                       dot15d4AddressField("", 0, adjust=lambda pkt, x: 8),
                       count_from=lambda pkt: pkt.pa_num_long),
        # TODO beacon payload
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Beacon ( %Dot15d4Beacon.src_panid%:%Dot15d4Beacon.src_addr% ) assocPermit(%Dot15d4Beacon.sf_assocpermit%) panCoord(%Dot15d4Beacon.sf_pancoord%)")  # noqa: E501


class Dot15d4Cmd(Packet):
    name = "802.15.4 Command"
    fields_desc = [
        XLEShortField("dest_panid", 0xFFFF),
        # Users should correctly set the dest_addr field. By default is 0x0 for construction to work.  # noqa: E501
        dot15d4AddressField("dest_addr", 0x0, length_of="fcf_destaddrmode"),
        ConditionalField(XLEShortField("src_panid", 0x0), \
                         lambda pkt:util_srcpanid_present(pkt)),
        ConditionalField(dot15d4AddressField("src_addr", None,
                         length_of="fcf_srcaddrmode"),
                         lambda pkt:pkt.underlayer.getfieldval("fcf_srcaddrmode") != 0),  # noqa: E501
        # Security field present if fcf_security == True
        ConditionalField(PacketField("aux_sec_header", Dot15d4AuxSecurityHeader(), Dot15d4AuxSecurityHeader),  # noqa: E501
                         lambda pkt:pkt.underlayer.getfieldval("fcf_security") is True),  # noqa: E501
        ByteEnumField("cmd_id", 0, {
            1: "AssocReq",  # Association request
            2: "AssocResp",  # Association response
            3: "DisassocNotify",  # Disassociation notification
            4: "DataReq",  # Data request
            5: "PANIDConflictNotify",  # PAN ID conflict notification
            6: "OrphanNotify",  # Orphan notification
            7: "BeaconReq",  # Beacon request
            8: "CoordRealign",  # coordinator realignment
            9: "GTSReq"  # GTS request
            # 0x0a - 0xff reserved
        }),
        # TODO command payload
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Command %Dot15d4Cmd.cmd_id% ( %Dot15dCmd.src_panid%:%Dot15d4Cmd.src_addr% -> %Dot15d4Cmd.dest_panid%:%Dot15d4Cmd.dest_addr% )")  # noqa: E501

    # command frame payloads are complete: DataReq, PANIDConflictNotify, OrphanNotify, BeaconReq don't have any payload  # noqa: E501
    # Although BeaconReq can have an optional ZigBee Beacon payload (implemented in ZigBeeBeacon)  # noqa: E501
    def guess_payload_class(self, payload):
        if self.cmd_id == 1:
            return Dot15d4CmdAssocReq
        elif self.cmd_id == 2:
            return Dot15d4CmdAssocResp
        elif self.cmd_id == 3:
            return Dot15d4CmdDisassociation
        elif self.cmd_id == 8:
            return Dot15d4CmdCoordRealign
        elif self.cmd_id == 9:
            return Dot15d4CmdGTSReq
        else:
            return Packet.guess_payload_class(self, payload)


class Dot15d4CmdCoordRealign(Packet):
    name = "802.15.4 Coordinator Realign Command"
    fields_desc = [
        # PAN Identifier (2 octets)
        XLEShortField("panid", 0xFFFF),
        # Coordinator Short Address (2 octets)
        XLEShortField("coord_address", 0x0000),
        # Logical Channel (1 octet): the logical channel that the coordinator intends to use for all future communications  # noqa: E501
        ByteField("channel", 0),
        # Short Address (2 octets)
        XLEShortField("dev_address", 0xFFFF),
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Coordinator Realign Payload ( PAN ID: %Dot15dCmdCoordRealign.pan_id% : channel %Dot15d4CmdCoordRealign.channel% )")  # noqa: E501

    def guess_payload_class(self, payload):
        if len(payload) == 1:
            return Dot15d4CmdCoordRealignPage
        else:
            return Packet.guess_payload_class(self, payload)


class Dot15d4CmdCoordRealignPage(Packet):
    name = "802.15.4 Coordinator Realign Page"
    fields_desc = [
        ByteField("channel_page", 0),
    ]


# Utility Functions #


def util_srcpanid_present(pkt):
    '''A source PAN ID is included if and only if both src addr mode != 0 and PAN ID Compression in FCF == 0'''  # noqa: E501
    if (pkt.underlayer.getfieldval("fcf_srcaddrmode") != 0) and (pkt.underlayer.getfieldval("fcf_panidcompress") == 0):  # noqa: E501
        return True
    else:
        return False


class Dot15d4CmdAssocReq(Packet):
    name = "802.15.4 Association Request Payload"
    fields_desc = [
        BitField("allocate_address", 0, 1),  # Allocate Address
        BitField("security_capability", 0, 1),  # Security Capability
        BitField("reserved2", 0, 1),  # bit 5 is reserved
        BitField("reserved1", 0, 1),  # bit 4 is reserved
        BitField("receiver_on_when_idle", 0, 1),  # Receiver On When Idle
        BitField("power_source", 0, 1),  # Power Source
        BitField("device_type", 0, 1),  # Device Type
        BitField("alternate_pan_coordinator", 0, 1),  # Alternate PAN Coordinator  # noqa: E501
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Association Request Payload ( Alt PAN Coord: %Dot15d4CmdAssocReq.alternate_pan_coordinator% Device Type: %Dot15d4CmdAssocReq.device_type% )")  # noqa: E501


class Dot15d4CmdAssocResp(Packet):
    name = "802.15.4 Association Response Payload"
    fields_desc = [
        XLEShortField("short_address", 0xFFFF),  # Address assigned to device from coordinator (0xFFFF == none)  # noqa: E501
        # Association Status
        # 0x00 == successful
        # 0x01 == PAN at capacity
        # 0x02 == PAN access denied
        # 0x03 - 0x7f == Reserved
        # 0x80 - 0xff == Reserved for MAC primitive enumeration values
        ByteEnumField("association_status", 0x00, {0: 'successful', 1: 'PAN_at_capacity', 2: 'PAN_access_denied'}),  # noqa: E501
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Association Response Payload ( Association Status: %Dot15d4CmdAssocResp.association_status% Assigned Address: %Dot15d4CmdAssocResp.short_address% )")  # noqa: E501


class Dot15d4CmdDisassociation(Packet):
    name = "802.15.4 Disassociation Notification Payload"
    fields_desc = [
        # Disassociation Reason
        # 0x00 == Reserved
        # 0x01 == The coordinator wishes the device to leave the PAN
        # 0x02 == The device wishes to leave the PAN
        # 0x03 - 0x7f == Reserved
        # 0x80 - 0xff == Reserved for MAC primitive enumeration values
        ByteEnumField("disassociation_reason", 0x02, {1: 'coord_wishes_device_to_leave', 2: 'device_wishes_to_leave'}),  # noqa: E501
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 Disassociation Notification Payload ( Disassociation Reason %Dot15d4CmdDisassociation.disassociation_reason% )")  # noqa: E501


class Dot15d4CmdGTSReq(Packet):
    name = "802.15.4 GTS request command"
    fields_desc = [
        # GTS Characteristics field (1 octet)
        # Reserved (bits 6-7)
        BitField("reserved", 0, 2),
        # Characteristics Type (bit 5)
        BitField("charact_type", 0, 1),
        # GTS Direction (bit 4)
        BitField("gts_dir", 0, 1),
        # GTS Length (bits 0-3)
        BitField("gts_len", 0, 4),
    ]

    def mysummary(self):
        return self.sprintf("802.15.4 GTS Request Command ( %Dot15d4CmdGTSReq.gts_len% : %Dot15d4CmdGTSReq.gts_dir% )")  # noqa: E501


# PAN ID conflict notification command frame is not necessary, only Dot15d4Cmd with cmd_id = 5 ("PANIDConflictNotify")  # noqa: E501
# Orphan notification command not necessary, only Dot15d4Cmd with cmd_id = 6 ("OrphanNotify")  # noqa: E501

# Bindings #
bind_layers(Dot15d4, Dot15d4Beacon, fcf_frametype=0)
bind_layers(Dot15d4, Dot15d4Data, fcf_frametype=1)
bind_layers(Dot15d4, Dot15d4Ack, fcf_frametype=2)
bind_layers(Dot15d4, Dot15d4Cmd, fcf_frametype=3)

# DLT Types #
conf.l2types.register(DLT_IEEE802_15_4_WITHFCS, Dot15d4FCS)
conf.l2types.register(DLT_IEEE802_15_4_NOFCS, Dot15d4)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
SMB 1 / 2 Client Automaton


.. note::
    You will find more complete documentation for this layer over at
    `SMB <https://scapy.readthedocs.io/en/latest/layers/smb.html#client>`_
"""

import io
import os
import pathlib
import socket
import time
import threading

from scapy.automaton import ATMT, Automaton, ObjectPipe
from scapy.config import conf
from scapy.error import Scapy_Exception
from scapy.fields import UTCTimeField
from scapy.supersocket import SuperSocket
from scapy.utils import (
    CLIUtil,
    pretty_list,
    human_size,
)
from scapy.volatile import RandUUID

from scapy.layers.dcerpc import NDRUnion, find_dcerpc_interface
from scapy.layers.gssapi import (
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    GSS_C_FLAGS,
)
from scapy.layers.msrpce.raw.ms_srvs import (
    LPSHARE_ENUM_STRUCT,
    NetrShareEnum_Request,
    NetrShareEnum_Response,
    SHARE_INFO_1_CONTAINER,
)
from scapy.layers.ntlm import (
    NTLMSSP,
)
from scapy.layers.smb import (
    SMBNegotiate_Request,
    SMBNegotiate_Response_Extended_Security,
    SMBNegotiate_Response_Security,
    SMBSession_Null,
    SMBSession_Setup_AndX_Request,
    SMBSession_Setup_AndX_Request_Extended_Security,
    SMBSession_Setup_AndX_Response,
    SMBSession_Setup_AndX_Response_Extended_Security,
    SMB_Dialect,
    SMB_Header,
)
from scapy.layers.smb2 import (
    DirectTCP,
    FileAllInformation,
    FileIdBothDirectoryInformation,
    SECURITY_DESCRIPTOR,
    SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2,
    SMB2_CREATE_REQUEST_LEASE,
    SMB2_CREATE_REQUEST_LEASE_V2,
    SMB2_Change_Notify_Request,
    SMB2_Change_Notify_Response,
    SMB2_Close_Request,
    SMB2_Close_Response,
    SMB2_Create_Context,
    SMB2_Create_Request,
    SMB2_Create_Response,
    SMB2_ENCRYPTION_CIPHERS,
    SMB2_Encryption_Capabilities,
    SMB2_Error_Response,
    SMB2_Header,
    SMB2_IOCTL_Request,
    SMB2_IOCTL_Response,
    SMB2_Negotiate_Context,
    SMB2_Negotiate_Protocol_Request,
    SMB2_Negotiate_Protocol_Response,
    SMB2_Netname_Negotiate_Context_ID,
    SMB2_Preauth_Integrity_Capabilities,
    SMB2_Query_Directory_Request,
    SMB2_Query_Directory_Response,
    SMB2_Query_Info_Request,
    SMB2_Query_Info_Response,
    SMB2_Read_Request,
    SMB2_Read_Response,
    SMB2_SIGNING_ALGORITHMS,
    SMB2_Session_Setup_Request,
    SMB2_Session_Setup_Response,
    SMB2_Signing_Capabilities,
    SMB2_Tree_Connect_Request,
    SMB2_Tree_Connect_Response,
    SMB2_Tree_Disconnect_Request,
    SMB2_Tree_Disconnect_Response,
    SMB2_Write_Request,
    SMB2_Write_Response,
    SMBStreamSocket,
    SMB_DIALECTS,
    SRVSVC_SHARE_TYPES,
    STATUS_ERREF,
)
from scapy.layers.spnego import SPNEGOSSP


class SMB_Client(Automaton):
    """
    SMB client automaton

    :param sock: the SMBStreamSocket to use
    :param ssp: the SSP to use

    All other options (in caps) are optional, and SMB specific:

    :param REQUIRE_SIGNATURE: set 'Require Signature'
    :param REQUIRE_ENCRYPTION: set 'Requite Encryption'
    :param MIN_DIALECT: minimum SMB dialect. Defaults to 0x0202 (2.0.2)
    :param MAX_DIALECT: maximum SMB dialect. Defaults to 0x0311 (3.1.1)
    :param DIALECTS: list of supported SMB2 dialects.
                     Constructed from MIN_DIALECT, MAX_DIALECT otherwise.
    """

    port = 445
    cls = DirectTCP

    def __init__(self, sock, ssp=None, *args, **kwargs):
        # Various SMB client arguments
        self.EXTENDED_SECURITY = kwargs.pop("EXTENDED_SECURITY", True)
        self.USE_SMB1 = kwargs.pop("USE_SMB1", False)
        self.REQUIRE_SIGNATURE = kwargs.pop("REQUIRE_SIGNATURE", None)
        self.REQUIRE_ENCRYPTION = kwargs.pop("REQUIRE_ENCRYPTION", False)
        self.RETRY = kwargs.pop("RETRY", 0)  # optionally: retry n times session setup
        self.SMB2 = kwargs.pop("SMB2", False)  # optionally: start directly in SMB2
        self.HOST = kwargs.pop("HOST", "")
        # Store supported dialects
        if "DIALECTS" in kwargs:
            self.DIALECTS = kwargs.pop("DIALECTS")
        else:
            MIN_DIALECT = kwargs.pop("MIN_DIALECT", 0x0202)
            self.MAX_DIALECT = kwargs.pop("MAX_DIALECT", 0x0311)
            self.DIALECTS = sorted(
                [
                    x
                    for x in [0x0202, 0x0210, 0x0300, 0x0302, 0x0311]
                    if x >= MIN_DIALECT and x <= self.MAX_DIALECT
                ]
            )
        # Internal Session information
        self.ErrorStatus = None
        self.NegotiateCapabilities = None
        self.GUID = RandUUID()._fix()
        self.SequenceWindow = (0, 0)  # keep track of allowed MIDs
        if ssp is None:
            # We got no SSP. Assuming the server allows anonymous
            ssp = SPNEGOSSP(
                [
                    NTLMSSP(
                        UPN="guest",
                        HASHNT=b"",
                    )
                ]
            )
        # Initialize
        kwargs["sock"] = sock
        Automaton.__init__(
            self,
            *args,
            **kwargs,
        )
        if self.is_atmt_socket:
            self.smb_sock_ready = threading.Event()
        # Set session options
        self.session.ssp = ssp
        self.session.SigningRequired = (
            self.REQUIRE_SIGNATURE if self.REQUIRE_SIGNATURE is not None else bool(ssp)
        )
        self.session.Dialect = self.MAX_DIALECT

    @classmethod
    def from_tcpsock(cls, sock, **kwargs):
        return cls.smblink(
            None,
            SMBStreamSocket(sock, DirectTCP),
            **kwargs,
        )

    @property
    def session(self):
        # session shorthand
        return self.sock.session

    def send(self, pkt):
        # Calculate what CreditCharge to send.
        if self.session.Dialect > 0x0202 and isinstance(pkt.payload, SMB2_Header):
            # [MS-SMB2] sect 3.2.4.1.5
            typ = type(pkt.payload.payload)
            if typ is SMB2_Negotiate_Protocol_Request:
                # See [MS-SMB2] 3.2.4.1.2 note
                pkt.CreditCharge = 0
            elif typ in [
                SMB2_Read_Request,
                SMB2_Write_Request,
                SMB2_IOCTL_Request,
                SMB2_Query_Directory_Request,
                SMB2_Change_Notify_Request,
                SMB2_Query_Info_Request,
            ]:
                # [MS-SMB2] 3.1.5.2
                # "For READ, WRITE, IOCTL, and QUERY_DIRECTORY requests"
                # "CHANGE_NOTIFY, QUERY_INFO, or SET_INFO"
                if typ == SMB2_Read_Request:
                    Length = pkt.payload.Length
                elif typ == SMB2_Write_Request:
                    Length = len(pkt.payload.Data)
                elif typ == SMB2_IOCTL_Request:
                    # [MS-SMB2] 3.3.5.15
                    Length = max(len(pkt.payload.Input), pkt.payload.MaxOutputResponse)
                elif typ in [
                    SMB2_Query_Directory_Request,
                    SMB2_Change_Notify_Request,
                    SMB2_Query_Info_Request,
                ]:
                    Length = pkt.payload.OutputBufferLength
                else:
                    raise RuntimeError("impossible case")
                pkt.CreditCharge = 1 + (Length - 1) // 65536
            else:
                # "For all other requests, the client MUST set CreditCharge to 1"
                pkt.CreditCharge = 1
            # [MS-SMB2] 3.2.4.1.2
            pkt.CreditRequest = pkt.CreditCharge + 1  # this code is a bit lazy
        # Get first available message ID: [MS-SMB2] 3.2.4.1.3 and 3.2.4.1.5
        pkt.MID = self.SequenceWindow[0]
        return super(SMB_Client, self).send(pkt)

    @ATMT.state(initial=1)
    def BEGIN(self):
        pass

    @ATMT.condition(BEGIN)
    def continue_smb2(self):
        if self.SMB2:  # Directly started in SMB2
            self.smb_header = DirectTCP() / SMB2_Header(PID=0xFEFF)
            raise self.SMB2_NEGOTIATE()

    @ATMT.condition(BEGIN, prio=1)
    def send_negotiate(self):
        raise self.SENT_NEGOTIATE()

    @ATMT.action(send_negotiate)
    def on_negotiate(self):
        # [MS-SMB2] sect 3.2.4.2.2.1 - Multi-Protocol Negotiate
        self.smb_header = DirectTCP() / SMB_Header(
            Flags2=(
                "LONG_NAMES+EAS+NT_STATUS+UNICODE+"
                "SMB_SECURITY_SIGNATURE+EXTENDED_SECURITY"
            ),
            TID=0xFFFF,
            PIDLow=0xFEFF,
            UID=0,
            MID=0,
        )
        if self.EXTENDED_SECURITY:
            self.smb_header.Flags2 += "EXTENDED_SECURITY"
        pkt = self.smb_header.copy() / SMBNegotiate_Request(
            Dialects=[
                SMB_Dialect(DialectString=x)
                for x in [
                    "PC NETWORK PROGRAM 1.0",
                    "LANMAN1.0",
                    "Windows for Workgroups 3.1a",
                    "LM1.2X002",
                    "LANMAN2.1",
                    "NT LM 0.12",
                ]
                + (["SMB 2.002", "SMB 2.???"] if not self.USE_SMB1 else [])
            ],
        )
        if not self.EXTENDED_SECURITY:
            pkt.Flags2 -= "EXTENDED_SECURITY"
        pkt[SMB_Header].Flags2 = (
            pkt[SMB_Header].Flags2
            - "SMB_SECURITY_SIGNATURE"
            + "SMB_SECURITY_SIGNATURE_REQUIRED+IS_LONG_NAME"
        )
        self.send(pkt)

    @ATMT.state()
    def SENT_NEGOTIATE(self):
        pass

    @ATMT.state()
    def SMB2_NEGOTIATE(self):
        pass

    @ATMT.condition(SMB2_NEGOTIATE)
    def send_negotiate_smb2(self):
        raise self.SENT_NEGOTIATE()

    @ATMT.action(send_negotiate_smb2)
    def on_negotiate_smb2(self):
        # [MS-SMB2] sect 3.2.4.2.2.2 - SMB2-Only Negotiate
        pkt = self.smb_header.copy() / SMB2_Negotiate_Protocol_Request(
            Dialects=self.DIALECTS,
            SecurityMode=(
                "SIGNING_ENABLED+SIGNING_REQUIRED"
                if self.session.SigningRequired
                else "SIGNING_ENABLED"
            ),
        )
        if self.MAX_DIALECT >= 0x0210:
            # "If the client implements the SMB 2.1 or SMB 3.x dialect, ClientGuid
            # MUST be set to the global ClientGuid value"
            pkt.ClientGUID = self.GUID
        # Capabilities: same as [MS-SMB2] 3.3.5.4
        self.NegotiateCapabilities = "+".join(
            [
                "DFS",
                "LEASING",
                "LARGE_MTU",
            ]
        )
        if self.MAX_DIALECT >= 0x0300:
            # "if Connection.Dialect belongs to the SMB 3.x dialect family ..."
            self.NegotiateCapabilities += "+" + "+".join(
                [
                    "MULTI_CHANNEL",
                    "PERSISTENT_HANDLES",
                    "DIRECTORY_LEASING",
                    "ENCRYPTION",
                ]
            )
        if self.MAX_DIALECT >= 0x0311:
            # "If the client implements the SMB 3.1.1 dialect, it MUST do"
            pkt.NegotiateContexts = [
                SMB2_Negotiate_Context()
                / SMB2_Preauth_Integrity_Capabilities(
                    # As for today, no other hash algorithm is described by the spec
                    HashAlgorithms=["SHA-512"],
                    Salt=self.session.Salt,
                ),
                SMB2_Negotiate_Context()
                / SMB2_Encryption_Capabilities(
                    Ciphers=self.session.SupportedCipherIds,
                ),
                # TODO support compression and RDMA
                SMB2_Negotiate_Context()
                / SMB2_Netname_Negotiate_Context_ID(
                    NetName=self.HOST,
                ),
                SMB2_Negotiate_Context()
                / SMB2_Signing_Capabilities(
                    SigningAlgorithms=self.session.SupportedSigningAlgorithmIds,
                ),
            ]
        pkt.Capabilities = self.NegotiateCapabilities
        # Send
        self.send(pkt)
        # If required, compute sessions
        self.session.computeSMBConnectionPreauth(
            bytes(pkt[SMB2_Header]),  # nego request
        )

    @ATMT.receive_condition(SENT_NEGOTIATE)
    def receive_negotiate_response(self, pkt):
        if (
            SMBNegotiate_Response_Extended_Security in pkt
            or SMB2_Negotiate_Protocol_Response in pkt
        ):
            # Extended SMB1 / SMB2
            try:
                ssp_blob = pkt.SecurityBlob  # eventually SPNEGO server initiation
            except AttributeError:
                ssp_blob = None
            if (
                SMB2_Negotiate_Protocol_Response in pkt
                and pkt.DialectRevision & 0xFF == 0xFF
            ):
                # Version is SMB X.???
                # [MS-SMB2] 3.2.5.2
                # If the DialectRevision field in the SMB2 NEGOTIATE Response is
                # 0x02FF ... the client MUST allocate sequence number 1 from
                # Connection.SequenceWindow, and MUST set MessageId field of the
                # SMB2 header to 1.
                self.SequenceWindow = (1, 1)
                self.smb_header = DirectTCP() / SMB2_Header(PID=0xFEFF, MID=1)
                self.SMB2 = True  # We're now using SMB2 to talk to the server
                raise self.SMB2_NEGOTIATE()
            else:
                if SMB2_Negotiate_Protocol_Response in pkt:
                    # SMB2 was negotiated !
                    self.session.Dialect = pkt.DialectRevision
                    # If required, compute sessions
                    self.session.computeSMBConnectionPreauth(
                        bytes(pkt[SMB2_Header]),  # nego response
                    )
                    # Process max sizes
                    self.session.MaxReadSize = pkt.MaxReadSize
                    self.session.MaxTransactionSize = pkt.MaxTransactionSize
                    self.session.MaxWriteSize = pkt.MaxWriteSize
                    # Process SecurityMode
                    if pkt.SecurityMode.SIGNING_REQUIRED:
                        self.session.SigningRequired = True
                    # Process capabilities
                    if self.session.Dialect >= 0x0300:
                        self.session.SupportsEncryption = pkt.Capabilities.ENCRYPTION
                    # Process NegotiateContext
                    if self.session.Dialect >= 0x0311 and pkt.NegotiateContextsCount:
                        for ngctx in pkt.NegotiateContexts:
                            if ngctx.ContextType == 0x0002:
                                # SMB2_ENCRYPTION_CAPABILITIES
                                if ngctx.Ciphers[0] != 0:
                                    self.session.CipherId = SMB2_ENCRYPTION_CIPHERS[
                                        ngctx.Ciphers[0]
                                    ]
                                    self.session.SupportsEncryption = True
                            elif ngctx.ContextType == 0x0008:
                                # SMB2_SIGNING_CAPABILITIES
                                self.session.SigningAlgorithmId = (
                                    SMB2_SIGNING_ALGORITHMS[ngctx.SigningAlgorithms[0]]
                                )
                    if self.REQUIRE_ENCRYPTION and not self.session.SupportsEncryption:
                        self.ErrorStatus = "NEGOTIATE FAILURE: encryption."
                        raise self.NEGO_FAILED()
                self.update_smbheader(pkt)
                raise self.NEGOTIATED(ssp_blob)
        elif SMBNegotiate_Response_Security in pkt:
            # Non-extended SMB1
            # Never tested. FIXME. probably broken
            raise self.NEGOTIATED(pkt.Challenge)

    @ATMT.state(final=1)
    def NEGO_FAILED(self):
        self.smb_sock_ready.set()

    @ATMT.state()
    def NEGOTIATED(self, ssp_blob=None):
        # Negotiated ! We now know the Dialect
        if self.session.Dialect > 0x0202:
            # [MS-SMB2] sect 3.2.5.1.4
            self.smb_header.CreditRequest = 1
        # Begin session establishment
        ssp_tuple = self.session.ssp.GSS_Init_sec_context(
            self.session.sspcontext,
            input_token=ssp_blob,
            target_name="cifs/" + self.HOST if self.HOST else None,
            req_flags=(
                GSS_C_FLAGS.GSS_C_MUTUAL_FLAG
                | (GSS_C_FLAGS.GSS_C_INTEG_FLAG if self.session.SigningRequired else 0)
            ),
        )
        return ssp_tuple

    def update_smbheader(self, pkt):
        """
        Called when receiving a SMB2 packet to update the current smb_header
        """
        # Some values should not be updated when ASYNC
        if not pkt.Flags.SMB2_FLAGS_ASYNC_COMMAND:
            # Update IDs
            self.smb_header.SessionId = pkt.SessionId
            self.smb_header.TID = pkt.TID
            self.smb_header.PID = pkt.PID
        # [MS-SMB2] 3.2.5.1.4
        self.SequenceWindow = (
            self.SequenceWindow[0] + max(pkt.CreditCharge, 1),
            self.SequenceWindow[1] + pkt.CreditRequest,
        )

    # DEV: add a condition on NEGOTIATED with prio=0

    @ATMT.condition(NEGOTIATED, prio=1)
    def should_send_session_setup_request(self, ssp_tuple):
        _, _, status = ssp_tuple
        if status not in [GSS_S_COMPLETE, GSS_S_CONTINUE_NEEDED]:
            raise ValueError("Internal error: the SSP completed with an error.")
        raise self.SENT_SESSION_REQUEST().action_parameters(ssp_tuple)

    @ATMT.state()
    def SENT_SESSION_REQUEST(self):
        pass

    @ATMT.action(should_send_session_setup_request)
    def send_setup_session_request(self, ssp_tuple):
        self.session.sspcontext, token, status = ssp_tuple
        if self.SMB2 and status == GSS_S_CONTINUE_NEEDED:
            # New session: force 0
            self.SessionId = 0
        if self.SMB2 or self.EXTENDED_SECURITY:
            # SMB1 extended / SMB2
            if self.SMB2:
                # SMB2
                pkt = self.smb_header.copy() / SMB2_Session_Setup_Request(
                    Capabilities="DFS",
                    SecurityMode=(
                        "SIGNING_ENABLED+SIGNING_REQUIRED"
                        if self.session.SigningRequired
                        else "SIGNING_ENABLED"
                    ),
                )
            else:
                # SMB1 extended
                pkt = (
                    self.smb_header.copy()
                    / SMBSession_Setup_AndX_Request_Extended_Security(
                        ServerCapabilities=(
                            "UNICODE+NT_SMBS+STATUS32+LEVEL_II_OPLOCKS+"
                            "DYNAMIC_REAUTH+EXTENDED_SECURITY"
                        ),
                        NativeOS=b"",
                        NativeLanMan=b"",
                    )
                )
            pkt.SecurityBlob = token
        else:
            # Non-extended security.
            pkt = self.smb_header.copy() / SMBSession_Setup_AndX_Request(
                ServerCapabilities="UNICODE+NT_SMBS+STATUS32+LEVEL_II_OPLOCKS",
                NativeOS=b"",
                NativeLanMan=b"",
                OEMPassword=b"\0" * 24,
                UnicodePassword=token,
            )
        self.send(pkt)
        if self.SMB2:
            # If required, compute sessions
            self.session.computeSMBSessionPreauth(
                bytes(pkt[SMB2_Header]),  # session request
            )

    @ATMT.receive_condition(SENT_SESSION_REQUEST)
    def receive_session_setup_response(self, pkt):
        if (
            SMBSession_Null in pkt
            or SMBSession_Setup_AndX_Response_Extended_Security in pkt
            or SMBSession_Setup_AndX_Response in pkt
        ):
            # SMB1
            if SMBSession_Null in pkt:
                # Likely an error
                raise self.NEGOTIATED()
        # Logging
        if pkt.Status != 0 and pkt.Status != 0xC0000016:
            # Not SUCCESS nor MORE_PROCESSING_REQUIRED: log
            self.ErrorStatus = pkt.sprintf("%SMB2_Header.Status%")
            self.debug(
                lvl=1,
                msg=conf.color_theme.red(
                    pkt.sprintf("SMB Session Setup Response: %SMB2_Header.Status%")
                ),
            )
        if self.SMB2:
            self.update_smbheader(pkt)
        # Cases depending on the response packet
        if (
            SMBSession_Setup_AndX_Response_Extended_Security in pkt
            or SMB2_Session_Setup_Response in pkt
        ):
            # The server assigns us a SessionId
            self.smb_header.SessionId = pkt.SessionId
            # SMB1 extended / SMB2
            if pkt.Status == 0:  # Authenticated
                if SMB2_Session_Setup_Response in pkt:
                    # [MS-SMB2] sect 3.2.5.3.1
                    if pkt.SessionFlags.IS_GUEST:
                        # "If the security subsystem indicates that the session
                        # was established by a guest user, Session.SigningRequired
                        # MUST be set to FALSE and Session.IsGuest MUST be set to TRUE."
                        self.session.IsGuest = True
                        self.session.SigningRequired = False
                    elif self.session.Dialect >= 0x0300:
                        if pkt.SessionFlags.ENCRYPT_DATA or self.REQUIRE_ENCRYPTION:
                            self.session.EncryptData = True
                            self.session.SigningRequired = False
                raise self.AUTHENTICATED(pkt.SecurityBlob)
            else:
                if SMB2_Header in pkt:
                    # If required, compute sessions
                    self.session.computeSMBSessionPreauth(
                        bytes(pkt[SMB2_Header]),  # session response
                    )
                # Ongoing auth
                raise self.NEGOTIATED(pkt.SecurityBlob)
        elif SMBSession_Setup_AndX_Response_Extended_Security in pkt:
            # SMB1 non-extended
            pass
        elif SMB2_Error_Response in pkt:
            # Authentication failure
            self.session.sspcontext.clifailure()
            # Reset Session preauth (SMB 3.1.1)
            self.session.SessionPreauthIntegrityHashValue = None
            if not self.RETRY:
                raise self.AUTH_FAILED()
            self.debug(lvl=2, msg="RETRY: %s" % self.RETRY)
            self.RETRY -= 1
            raise self.NEGOTIATED()

    @ATMT.state(final=1)
    def AUTH_FAILED(self):
        self.smb_sock_ready.set()

    @ATMT.state()
    def AUTHENTICATED(self, ssp_blob=None):
        self.session.sspcontext, _, status = self.session.ssp.GSS_Init_sec_context(
            self.session.sspcontext,
            input_token=ssp_blob,
            target_name="cifs/" + self.HOST if self.HOST else None,
        )
        if status != GSS_S_COMPLETE:
            raise ValueError("Internal error: the SSP completed with an error.")
        # Authentication was successful
        self.session.computeSMBSessionKeys(IsClient=True)

    # DEV: add a condition on AUTHENTICATED with prio=0

    @ATMT.condition(AUTHENTICATED, prio=1)
    def authenticated_post_actions(self):
        raise self.SOCKET_BIND()

    # Plain SMB Socket

    @ATMT.state()
    def SOCKET_BIND(self):
        self.smb_sock_ready.set()

    @ATMT.condition(SOCKET_BIND)
    def start_smb_socket(self):
        raise self.SOCKET_MODE_SMB()

    @ATMT.state()
    def SOCKET_MODE_SMB(self):
        pass

    @ATMT.receive_condition(SOCKET_MODE_SMB)
    def incoming_data_received_smb(self, pkt):
        raise self.SOCKET_MODE_SMB().action_parameters(pkt)

    @ATMT.action(incoming_data_received_smb)
    def receive_data_smb(self, pkt):
        resp = pkt[SMB2_Header].payload
        if isinstance(resp, SMB2_Error_Response):
            if pkt.Status == 0x00000103:  # STATUS_PENDING
                # answer is coming later.. just wait...
                return
            if pkt.Status == 0x0000010B:  # STATUS_NOTIFY_CLEANUP
                # this is a notify cleanup. ignore
                return
        self.update_smbheader(pkt)
        # Add the status to the response as metadata
        resp.NTStatus = pkt.sprintf("%SMB2_Header.Status%")
        self.oi.smbpipe.send(resp)

    @ATMT.ioevent(SOCKET_MODE_SMB, name="smbpipe", as_supersocket="smblink")
    def outgoing_data_received_smb(self, fd):
        raise self.SOCKET_MODE_SMB().action_parameters(fd.recv())

    @ATMT.action(outgoing_data_received_smb)
    def send_data(self, d):
        self.send(self.smb_header.copy() / d)


class SMB_SOCKET(SuperSocket):
    """
    Mid-level wrapper over SMB_Client.smblink that provides some basic SMB
    client functions, such as tree connect, directory query, etc.
    """

    def __init__(self, smbsock, use_ioctl=True, timeout=3):
        self.ins = smbsock
        self.timeout = timeout
        if not self.ins.atmt.smb_sock_ready.wait(timeout=timeout):
            # If we have a SSP, tell it we failed.
            if self.ins.atmt.session.sspcontext:
                self.ins.atmt.session.sspcontext.clifailure()
            raise TimeoutError(
                "The SMB handshake timed out ! (enable debug=1 for logs)"
            )
        if self.ins.atmt.ErrorStatus:
            raise Scapy_Exception(
                "SMB Session Setup failed: %s" % self.ins.atmt.ErrorStatus
            )

    @classmethod
    def from_tcpsock(cls, sock, **kwargs):
        """
        Wraps the tcp socket in a SMB_Client.smblink first, then into the
        SMB_SOCKET/SMB_RPC_SOCKET
        """
        return cls(
            use_ioctl=kwargs.pop("use_ioctl", True),
            timeout=kwargs.pop("timeout", 3),
            smbsock=SMB_Client.from_tcpsock(sock, **kwargs),
        )

    @property
    def session(self):
        return self.ins.atmt.session

    def set_TID(self, TID):
        """
        Set the TID (Tree ID).
        This can be called before sending a packet
        """
        self.ins.atmt.smb_header.TID = TID

    def get_TID(self):
        """
        Get the current TID from the underlying socket
        """
        return self.ins.atmt.smb_header.TID

    def tree_connect(self, name):
        """
        Send a TreeConnect request
        """
        resp = self.ins.sr1(
            SMB2_Tree_Connect_Request(
                Buffer=[
                    (
                        "Path",
                        "\\\\%s\\%s"
                        % (
                            self.session.sspcontext.ServerHostname,
                            name,
                        ),
                    )
                ]
            ),
            verbose=False,
            timeout=self.timeout,
        )
        if not resp:
            raise ValueError("TreeConnect timed out !")
        if SMB2_Tree_Connect_Response not in resp:
            raise ValueError("Failed TreeConnect ! %s" % resp.NTStatus)
        # [MS-SMB2] sect 3.2.5.5
        if self.session.Dialect >= 0x0300:
            if resp.ShareFlags.ENCRYPT_DATA and self.session.SupportsEncryption:
                self.session.TreeEncryptData = True
            else:
                self.session.TreeEncryptData = False
        return self.get_TID()

    def tree_disconnect(self):
        """
        Send a TreeDisconnect request
        """
        resp = self.ins.sr1(
            SMB2_Tree_Disconnect_Request(),
            verbose=False,
            timeout=self.timeout,
        )
        if not resp:
            raise ValueError("TreeDisconnect timed out !")
        if SMB2_Tree_Disconnect_Response not in resp:
            raise ValueError("Failed TreeDisconnect ! %s" % resp.NTStatus)

    def create_request(
        self,
        name,
        mode="r",
        type="pipe",
        extra_create_options=[],
        extra_desired_access=[],
    ):
        """
        Open a file/pipe by its name

        :param name: the name of the file or named pipe. e.g. 'srvsvc'
        """
        ShareAccess = []
        DesiredAccess = []
        # Common params depending on the access
        if "r" in mode:
            ShareAccess.append("FILE_SHARE_READ")
            DesiredAccess.extend(["FILE_READ_DATA", "FILE_READ_ATTRIBUTES"])
        if "w" in mode:
            ShareAccess.append("FILE_SHARE_WRITE")
            DesiredAccess.extend(["FILE_WRITE_DATA", "FILE_WRITE_ATTRIBUTES"])
        if "d" in mode:
            ShareAccess.append("FILE_SHARE_DELETE")
        # Params depending on the type
        FileAttributes = []
        CreateOptions = []
        CreateContexts = []
        CreateDisposition = "FILE_OPEN"
        if type == "folder":
            FileAttributes.append("FILE_ATTRIBUTE_DIRECTORY")
            CreateOptions.append("FILE_DIRECTORY_FILE")
        elif type in ["file", "pipe"]:
            CreateOptions = ["FILE_NON_DIRECTORY_FILE"]
            if "r" in mode:
                DesiredAccess.extend(["FILE_READ_EA", "READ_CONTROL", "SYNCHRONIZE"])
            if "w" in mode:
                CreateDisposition = "FILE_OVERWRITE_IF"
                DesiredAccess.append("FILE_WRITE_EA")
            if "d" in mode:
                DesiredAccess.append("DELETE")
                CreateOptions.append("FILE_DELETE_ON_CLOSE")
            if type == "file":
                FileAttributes.append("FILE_ATTRIBUTE_NORMAL")
        elif type:
            raise ValueError("Unknown type: %s" % type)
        # [MS-SMB2] 3.2.4.3.8
        RequestedOplockLevel = 0
        if self.session.Dialect >= 0x0300:
            RequestedOplockLevel = "SMB2_OPLOCK_LEVEL_LEASE"
        elif self.session.Dialect >= 0x0210 and type == "file":
            RequestedOplockLevel = "SMB2_OPLOCK_LEVEL_LEASE"
        # SMB 3.X
        if self.session.Dialect >= 0x0300 and type in ["file", "folder"]:
            CreateContexts.extend(
                [
                    # [SMB2] sect 3.2.4.3.5
                    SMB2_Create_Context(
                        Name=b"DH2Q",
                        Data=SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2(
                            CreateGuid=RandUUID()._fix()
                        ),
                    ),
                    # [SMB2] sect 3.2.4.3.9
                    SMB2_Create_Context(
                        Name=b"MxAc",
                    ),
                    # [SMB2] sect 3.2.4.3.10
                    SMB2_Create_Context(
                        Name=b"QFid",
                    ),
                    # [SMB2] sect 3.2.4.3.8
                    SMB2_Create_Context(
                        Name=b"RqLs",
                        Data=SMB2_CREATE_REQUEST_LEASE_V2(LeaseKey=RandUUID()._fix()),
                    ),
                ]
            )
        elif self.session.Dialect == 0x0210 and type == "file":
            CreateContexts.extend(
                [
                    # [SMB2] sect 3.2.4.3.8
                    SMB2_Create_Context(
                        Name=b"RqLs",
                        Data=SMB2_CREATE_REQUEST_LEASE(LeaseKey=RandUUID()._fix()),
                    ),
                ]
            )
        # Extra options
        if extra_create_options:
            CreateOptions.extend(extra_create_options)
        if extra_desired_access:
            DesiredAccess.extend(extra_desired_access)
        # Request
        resp = self.ins.sr1(
            SMB2_Create_Request(
                ImpersonationLevel="Impersonation",
                DesiredAccess="+".join(DesiredAccess),
                CreateDisposition=CreateDisposition,
                CreateOptions="+".join(CreateOptions),
                ShareAccess="+".join(ShareAccess),
                FileAttributes="+".join(FileAttributes),
                CreateContexts=CreateContexts,
                RequestedOplockLevel=RequestedOplockLevel,
                Name=name,
            ),
            verbose=0,
            timeout=self.timeout,
        )
        if not resp:
            raise ValueError("CreateRequest timed out !")
        if SMB2_Create_Response not in resp:
            raise ValueError("Failed CreateRequest ! %s" % resp.NTStatus)
        return resp[SMB2_Create_Response].FileId

    def close_request(self, FileId):
        """
        Close the FileId
        """
        pkt = SMB2_Close_Request(FileId=FileId)
        resp = self.ins.sr1(pkt, verbose=0, timeout=self.timeout)
        if not resp:
            raise ValueError("CloseRequest timed out !")
        if SMB2_Close_Response not in resp:
            raise ValueError("Failed CloseRequest ! %s" % resp.NTStatus)

    def read_request(self, FileId, Length, Offset=0):
        """
        Read request
        """
        resp = self.ins.sr1(
            SMB2_Read_Request(
                FileId=FileId,
                Length=Length,
                Offset=Offset,
            ),
            verbose=0,
            timeout=self.timeout,
        )
        if not resp:
            raise ValueError("ReadRequest timed out !")
        if SMB2_Read_Response not in resp:
            raise ValueError("Failed ReadRequest ! %s" % resp.NTStatus)
        return resp.Data

    def write_request(self, Data, FileId, Offset=0):
        """
        Write request
        """
        resp = self.ins.sr1(
            SMB2_Write_Request(
                FileId=FileId,
                Data=Data,
                Offset=Offset,
            ),
            verbose=0,
            timeout=self.timeout,
        )
        if not resp:
            raise ValueError("WriteRequest timed out !")
        if SMB2_Write_Response not in resp:
            raise ValueError("Failed WriteRequest ! %s" % resp.NTStatus)
        return resp.Count

    def query_directory(self, FileId, FileName="*"):
        """
        Query the Directory with FileId
        """
        results = []
        Flags = "SMB2_RESTART_SCANS"
        while True:
            pkt = SMB2_Query_Directory_Request(
                FileInformationClass="FileIdBothDirectoryInformation",
                FileId=FileId,
                FileName=FileName,
                Flags=Flags,
            )
            resp = self.ins.sr1(pkt, verbose=0, timeout=self.timeout)
            Flags = 0  # only the first one is RESTART_SCANS
            if not resp:
                raise ValueError("QueryDirectory timed out !")
            if SMB2_Error_Response in resp:
                break
            elif SMB2_Query_Directory_Response not in resp:
                raise ValueError("Failed QueryDirectory ! %s" % resp.NTStatus)
            res = FileIdBothDirectoryInformation(resp.Output)
            results.extend(
                [
                    (
                        x.FileName,
                        x.FileAttributes,
                        x.EndOfFile,
                        x.LastWriteTime,
                    )
                    for x in res.files
                ]
            )
        return results

    def query_info(self, FileId, InfoType, FileInfoClass, AdditionalInformation=0):
        """
        Query the Info
        """
        pkt = SMB2_Query_Info_Request(
            InfoType=InfoType,
            FileInfoClass=FileInfoClass,
            OutputBufferLength=65535,
            FileId=FileId,
            AdditionalInformation=AdditionalInformation,
        )
        resp = self.ins.sr1(pkt, verbose=0, timeout=self.timeout)
        if not resp:
            raise ValueError("QueryInfo timed out !")
        if SMB2_Query_Info_Response not in resp:
            raise ValueError("Failed QueryInfo ! %s" % resp.NTStatus)
        return resp.Output

    def changenotify(self, FileId):
        """
        Register change notify
        """
        pkt = SMB2_Change_Notify_Request(
            Flags="SMB2_WATCH_TREE",
            OutputBufferLength=65535,
            FileId=FileId,
            CompletionFilter=0x0FFF,
        )
        # we can wait forever, not a problem in this one
        resp = self.ins.sr1(pkt, verbose=0, chainCC=True)
        if SMB2_Change_Notify_Response not in resp:
            raise ValueError("Failed ChangeNotify ! %s" % resp.NTStatus)
        return resp.Output


class SMB_RPC_SOCKET(ObjectPipe, SMB_SOCKET):
    """
    Extends SMB_SOCKET (which is a wrapper over SMB_Client.smblink) to send
    DCE/RPC messages (bind, reqs, etc.)

    This is usable as a normal SuperSocket (sr1, etc.) and performs the
    wrapping of the DCE/RPC messages into SMB2_Write/Read packets.
    """

    def __init__(self, smbsock, use_ioctl=True, timeout=3):
        self.use_ioctl = use_ioctl
        ObjectPipe.__init__(self, "SMB_RPC_SOCKET")
        SMB_SOCKET.__init__(self, smbsock, timeout=timeout)

    def open_pipe(self, name):
        self.PipeFileId = self.create_request(name, mode="rw", type="pipe")

    def close_pipe(self):
        self.close_request(self.PipeFileId)
        self.PipeFileId = None

    def send(self, x):
        """
        Internal ObjectPipe function.
        """
        # Reminder: this class is an ObjectPipe, it's just a queue.

        # Detect if DCE/RPC is fragmented. Then we must use Read/Write
        is_frag = x.pfc_flags & 3 != 3

        if self.use_ioctl and not is_frag:
            # Use IOCTLRequest
            pkt = SMB2_IOCTL_Request(
                FileId=self.PipeFileId,
                Flags="SMB2_0_IOCTL_IS_FSCTL",
                CtlCode="FSCTL_PIPE_TRANSCEIVE",
            )
            pkt.Input = bytes(x)
            resp = self.ins.sr1(pkt, verbose=0)
            if SMB2_IOCTL_Response not in resp:
                raise ValueError("Failed reading IOCTL_Response ! %s" % resp.NTStatus)
            data = bytes(resp.Output)
            super(SMB_RPC_SOCKET, self).send(data)
            # Handle BUFFER_OVERFLOW (big DCE/RPC response)
            while resp.NTStatus == "STATUS_BUFFER_OVERFLOW" or data[3] & 2 != 2:
                # Retrieve DCE/RPC full size
                resp = self.ins.sr1(
                    SMB2_Read_Request(
                        FileId=self.PipeFileId,
                    ),
                    verbose=0,
                )
                data = resp.Data
                super(SMB_RPC_SOCKET, self).send(data)
        else:
            # Use WriteRequest/ReadRequest
            pkt = SMB2_Write_Request(
                FileId=self.PipeFileId,
            )
            pkt.Data = bytes(x)
            # We send the Write Request
            resp = self.ins.sr1(pkt, verbose=0)
            if SMB2_Write_Response not in resp:
                raise ValueError("Failed sending WriteResponse ! %s" % resp.NTStatus)
            # If fragmented, only read if it's the last.
            if is_frag and not x.pfc_flags.PFC_LAST_FRAG:
                return
            # We send a Read Request afterwards
            resp = self.ins.sr1(
                SMB2_Read_Request(
                    FileId=self.PipeFileId,
                ),
                verbose=0,
            )
            if SMB2_Read_Response not in resp:
                raise ValueError("Failed reading ReadResponse ! %s" % resp.NTStatus)
            super(SMB_RPC_SOCKET, self).send(resp.Data)
            # Handle fragmented response
            while resp.Data[3] & 2 != 2:  # PFC_LAST_FRAG not set
                # Retrieve DCE/RPC full size
                resp = self.ins.sr1(
                    SMB2_Read_Request(
                        FileId=self.PipeFileId,
                    ),
                    verbose=0,
                )
                super(SMB_RPC_SOCKET, self).send(resp.Data)

    def close(self):
        SMB_SOCKET.close(self)
        ObjectPipe.close(self)


@conf.commands.register
class smbclient(CLIUtil):
    r"""
    A simple SMB client CLI powered by Scapy

    :param target: can be a hostname, the IPv4 or the IPv6 to connect to
    :param UPN: the upn to use (DOMAIN/USER, DOMAIN\USER, USER@DOMAIN or USER)
    :param guest: use guest mode (over NTLM)
    :param ssp: if provided, use this SSP for auth.
    :param kerberos_required: require kerberos
    :param port: the TCP port. default 445
    :param password: if provided, used for auth
    :param HashNt: if provided, used for auth (NTLM)
    :param HashAes256Sha96: if provided, used for auth (Kerberos)
    :param HashAes128Sha96: if provided, used for auth (Kerberos)
    :param ST: if provided, the service ticket to use (Kerberos)
    :param KEY: if provided, the session key associated to the ticket (Kerberos)
    :param cli: CLI mode (default True). False to use for scripting

    Some additional SMB parameters are available under help(SMB_Client). Some of
    them include the following:

    :param REQUIRE_ENCRYPTION: requires encryption.
    """

    def __init__(
        self,
        target: str,
        UPN: str = None,
        password: str = None,
        guest: bool = False,
        kerberos_required: bool = False,
        HashNt: bytes = None,
        HashAes256Sha96: bytes = None,
        HashAes128Sha96: bytes = None,
        port: int = 445,
        timeout: int = 5,
        debug: int = 0,
        ssp=None,
        ST=None,
        KEY=None,
        cli=True,
        # SMB arguments
        REQUIRE_ENCRYPTION=False,
        **kwargs,
    ):
        if cli:
            self._depcheck()
        assert UPN or ssp or guest, "Either UPN, ssp or guest must be provided !"
        # Do we need to build a SSP?
        if ssp is None:
            # Create the SSP (only if not guest mode)
            if not guest:
                ssp = SPNEGOSSP.from_cli_arguments(
                    UPN=UPN,
                    target=target,
                    password=password,
                    HashNt=HashNt,
                    HashAes256Sha96=HashAes256Sha96,
                    HashAes128Sha96=HashAes128Sha96,
                    ST=ST,
                    KEY=KEY,
                    kerberos_required=kerberos_required,
                )
            else:
                # Guest mode
                ssp = None
        # Check if target is IPv4 or IPv6
        if ":" in target:
            family = socket.AF_INET6
        else:
            family = socket.AF_INET
        # Open socket
        sock = socket.socket(family, socket.SOCK_STREAM)
        # Configure socket for SMB:
        # - TCP KEEPALIVE, TCP_KEEPIDLE and TCP_KEEPINTVL. Against a Windows server this
        #   isn't necessary, but samba kills the socket VERY fast otherwise.
        # - set TCP_NODELAY to disable Nagle's algorithm (we're streaming data)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 10)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 10)
        # Timeout & connect
        sock.settimeout(timeout)
        if debug:
            print("Connecting to %s:%s" % (target, port))
        sock.connect((target, port))
        self.extra_create_options = []
        # Wrap with the automaton
        self.timeout = timeout
        kwargs.setdefault("HOST", target)
        self.sock = SMB_Client.from_tcpsock(
            sock,
            ssp=ssp,
            debug=debug,
            REQUIRE_ENCRYPTION=REQUIRE_ENCRYPTION,
            timeout=timeout,
            **kwargs,
        )
        try:
            # Wrap with SMB_SOCKET
            self.smbsock = SMB_SOCKET(self.sock, timeout=self.timeout)
            # Wait for either the atmt to fail, or the smb_sock_ready to timeout
            _t = time.time()
            while True:
                if self.sock.atmt.smb_sock_ready.is_set():
                    # yay
                    break
                if not self.sock.atmt.isrunning():
                    status = self.sock.atmt.get("Status")
                    raise Scapy_Exception(
                        "%s with status %s"
                        % (
                            self.sock.atmt.state.state,
                            STATUS_ERREF.get(status, hex(status)),
                        )
                    )
                if time.time() - _t > timeout:
                    self.sock.close()
                    raise TimeoutError("The SMB handshake timed out.")
                time.sleep(0.1)
        except Exception:
            # Something bad happened, end the socket/automaton
            self.sock.close()
            raise

        # For some usages, we will also need the RPC wrapper
        from scapy.layers.msrpce.rpcclient import DCERPC_Client

        self.rpcclient = DCERPC_Client.from_smblink(
            self.sock,
            ndr64=False,
            verb=bool(debug),
        )
        # We have a valid smb connection !
        print(
            "%s authentication successful using %s%s !"
            % (
                SMB_DIALECTS.get(
                    self.smbsock.session.Dialect,
                    "SMB %s" % self.smbsock.session.Dialect,
                ),
                repr(self.smbsock.session.sspcontext),
                " as GUEST" if self.smbsock.session.IsGuest else "",
            )
        )
        # Now define some variables for our CLI
        self.pwd = pathlib.PureWindowsPath("/")
        self.localpwd = pathlib.Path(".").resolve()
        self.current_tree = None
        self.ls_cache = {}  # cache the listing of the current directory
        self.sh_cache = []  # cache the shares
        # Start CLI
        if cli:
            self.loop(debug=debug)

    def ps1(self):
        return r"smb: \%s> " % self.normalize_path(self.pwd)

    def close(self):
        print("Connection closed")
        self.smbsock.close()

    def _require_share(self, silent=False):
        if self.current_tree is None:
            if not silent:
                print("No share selected ! Try 'shares' then 'use'.")
            return True

    def collapse_path(self, path):
        # the amount of pathlib.wtf you need to do to resolve .. on all platforms
        # is ridiculous
        return pathlib.PureWindowsPath(os.path.normpath(path.as_posix()))

    def normalize_path(self, path):
        """
        Normalize path for CIFS usage
        """
        return str(self.collapse_path(path)).lstrip("\\")

    @CLIUtil.addcommand()
    def shares(self):
        """
        List the shares available
        """
        # Poll cache
        if self.sh_cache:
            return self.sh_cache
        # It's an RPC
        self.rpcclient.open_smbpipe("srvsvc")
        self.rpcclient.bind(find_dcerpc_interface("srvsvc"))
        req = NetrShareEnum_Request(
            InfoStruct=LPSHARE_ENUM_STRUCT(
                Level=1,
                ShareInfo=NDRUnion(
                    tag=1,
                    value=SHARE_INFO_1_CONTAINER(Buffer=None),
                ),
            ),
            PreferedMaximumLength=0xFFFFFFFF,
            ndr64=self.rpcclient.ndr64,
        )
        resp = self.rpcclient.sr1_req(req, timeout=self.timeout)
        self.rpcclient.close_smbpipe()
        if not isinstance(resp, NetrShareEnum_Response):
            resp.show()
            raise ValueError("NetrShareEnum_Request failed !")
        results = []
        for share in resp.valueof("InfoStruct.ShareInfo.Buffer"):
            shi1_type = share.valueof("shi1_type") & 0x0FFFFFFF
            results.append(
                (
                    share.valueof("shi1_netname").decode(),
                    SRVSVC_SHARE_TYPES.get(shi1_type, shi1_type),
                    share.valueof("shi1_remark").decode(),
                )
            )
        self.sh_cache = results  # cache
        return results

    @CLIUtil.addoutput(shares)
    def shares_output(self, results):
        """
        Print the output of 'shares'
        """
        print(pretty_list(results, [("ShareName", "ShareType", "Comment")]))

    @CLIUtil.addcommand()
    def use(self, share):
        """
        Open a share
        """
        self.current_tree = self.smbsock.tree_connect(share)
        self.pwd = pathlib.PureWindowsPath("/")
        self.ls_cache.clear()

    @CLIUtil.addcomplete(use)
    def use_complete(self, share):
        """
        Auto-complete 'use'
        """
        return [
            x[0] for x in self.shares() if x[0].startswith(share) and x[0] != "IPC$"
        ]

    def _parsepath(self, arg, remote=True):
        """
        Parse a path. Returns the parent folder and file name
        """
        # Find parent directory if it exists
        elt = (pathlib.PureWindowsPath if remote else pathlib.Path)(arg)
        eltpar = (pathlib.PureWindowsPath if remote else pathlib.Path)(".")
        eltname = elt.name
        if arg.endswith("/") or arg.endswith("\\"):
            eltpar = elt
            eltname = ""
        elif elt.parent and elt.parent.name or elt.is_absolute():
            eltpar = elt.parent
        return eltpar, eltname

    def _fs_complete(self, arg, cond=None):
        """
        Return a listing of the remote files for completion purposes
        """
        if cond is None:
            cond = lambda _: True
        eltpar, eltname = self._parsepath(arg)
        # ls in that directory
        try:
            files = self.ls(parent=eltpar)
        except ValueError:
            return []
        return [
            str(eltpar / x[0])
            for x in files
            if (
                x[0].lower().startswith(eltname.lower())
                and x[0] not in [".", ".."]
                and cond(x[1])
            )
        ]

    def _dir_complete(self, arg):
        """
        Return a directories of remote files for completion purposes
        """
        results = self._fs_complete(
            arg,
            cond=lambda x: x.FILE_ATTRIBUTE_DIRECTORY,
        )
        if len(results) == 1 and results[0].startswith(arg):
            # skip through folders
            return [results[0] + "\\"]
        return results

    @CLIUtil.addcommand(spaces=True)
    def ls(self, parent=None):
        """
        List the files in the remote directory
        -t: sort by timestamp
        -S: sort by size
        -r: reverse while sorting
        """
        if self._require_share():
            return
        # Get pwd of the ls
        pwd = self.pwd
        if parent is not None:
            pwd /= parent
        pwd = self.normalize_path(pwd)
        # Poll the cache
        if self.ls_cache and pwd in self.ls_cache:
            return self.ls_cache[pwd]
        self.smbsock.set_TID(self.current_tree)
        # Open folder
        fileId = self.smbsock.create_request(
            pwd,
            type="folder",
            extra_create_options=self.extra_create_options,
        )
        # Query the folder
        files = self.smbsock.query_directory(fileId)
        # Close the folder
        self.smbsock.close_request(fileId)
        self.ls_cache[pwd] = files  # Store cache
        return files

    @CLIUtil.addoutput(ls)
    def ls_output(self, results, *, t=False, S=False, r=False):
        """
        Print the output of 'ls'
        """
        fld = UTCTimeField(
            "", None, fmt="<Q", epoch=[1601, 1, 1, 0, 0, 0], custom_scaling=1e7
        )
        if t:
            # Sort by time
            results.sort(key=lambda x: -x[3])
        if S:
            # Sort by size
            results.sort(key=lambda x: -x[2])
        if r:
            # Reverse sort
            results = results[::-1]
        results = [
            (
                x[0],
                "+".join(y.lstrip("FILE_ATTRIBUTE_") for y in str(x[1]).split("+")),
                human_size(x[2]),
                fld.i2repr(None, x[3]),
            )
            for x in results
        ]
        print(
            pretty_list(
                results,
                [("FileName", "FileAttributes", "EndOfFile", "LastWriteTime")],
                sortBy=None,
            )
        )

    @CLIUtil.addcomplete(ls)
    def ls_complete(self, folder):
        """
        Auto-complete ls
        """
        if self._require_share(silent=True):
            return []
        return self._dir_complete(folder)

    @CLIUtil.addcommand(spaces=True)
    def cd(self, folder):
        """
        Change the remote current directory
        """
        if self._require_share():
            return
        if not folder:
            # show mode
            return str(self.pwd)
        self.pwd /= folder
        self.pwd = self.collapse_path(self.pwd)
        self.ls_cache.clear()

    @CLIUtil.addcomplete(cd)
    def cd_complete(self, folder):
        """
        Auto-complete cd
        """
        if self._require_share(silent=True):
            return []
        return self._dir_complete(folder)

    def _lfs_complete(self, arg, cond):
        """
        Return a listing of local files for completion purposes
        """
        eltpar, eltname = self._parsepath(arg, remote=False)
        eltpar = self.localpwd / eltpar
        return [
            # trickery so that ../<TAB> works
            str(eltpar / x.name)
            for x in eltpar.resolve().glob("*")
            if (x.name.lower().startswith(eltname.lower()) and cond(x))
        ]

    @CLIUtil.addoutput(cd)
    def cd_output(self, result):
        """
        Print the output of 'cd'
        """
        if result:
            print(result)

    @CLIUtil.addcommand()
    def lls(self):
        """
        List the files in the local directory
        """
        return list(self.localpwd.glob("*"))

    @CLIUtil.addoutput(lls)
    def lls_output(self, results):
        """
        Print the output of 'lls'
        """
        results = [
            (
                x.name,
                human_size(stat.st_size),
                time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(stat.st_mtime)),
            )
            for x, stat in ((x, x.stat()) for x in results)
        ]
        print(
            pretty_list(results, [("FileName", "File Size", "Last Modification Time")])
        )

    @CLIUtil.addcommand(spaces=True)
    def lcd(self, folder):
        """
        Change the local current directory
        """
        if not folder:
            # show mode
            return str(self.localpwd)
        self.localpwd /= folder
        self.localpwd = self.localpwd.resolve()

    @CLIUtil.addcomplete(lcd)
    def lcd_complete(self, folder):
        """
        Auto-complete lcd
        """
        return self._lfs_complete(folder, lambda x: x.is_dir())

    @CLIUtil.addoutput(lcd)
    def lcd_output(self, result):
        """
        Print the output of 'lcd'
        """
        if result:
            print(result)

    def _get_file(self, file, fd):
        """
        Gets the file bytes from a remote host
        """
        # Get pwd of the ls
        fpath = self.pwd / file
        self.smbsock.set_TID(self.current_tree)
        # Open file
        fileId = self.smbsock.create_request(
            self.normalize_path(fpath),
            type="file",
            extra_create_options=[
                "FILE_SEQUENTIAL_ONLY",
            ]
            + self.extra_create_options,
        )
        # Get the file size
        info = FileAllInformation(
            self.smbsock.query_info(
                FileId=fileId,
                InfoType="SMB2_0_INFO_FILE",
                FileInfoClass="FileAllInformation",
            )
        )
        length = info.StandardInformation.EndOfFile
        offset = 0
        # Read the file
        while length:
            lengthRead = min(self.smbsock.session.MaxReadSize, length)
            fd.write(
                self.smbsock.read_request(fileId, Length=lengthRead, Offset=offset)
            )
            offset += lengthRead
            length -= lengthRead
        # Close the file
        self.smbsock.close_request(fileId)
        return offset

    def _send_file(self, fname, fd):
        """
        Send the file bytes to a remote host
        """
        # Get destination file
        fpath = self.pwd / fname
        self.smbsock.set_TID(self.current_tree)
        # Open file
        fileId = self.smbsock.create_request(
            self.normalize_path(fpath),
            type="file",
            mode="w",
            extra_create_options=self.extra_create_options,
        )
        # Send the file
        offset = 0
        while True:
            data = fd.read(self.smbsock.session.MaxWriteSize)
            if not data:
                # end of file
                break
            offset += self.smbsock.write_request(
                Data=data,
                FileId=fileId,
                Offset=offset,
            )
        # Close the file
        self.smbsock.close_request(fileId)
        return offset

    def _getr(self, directory, _root, _verb=True):
        """
        Internal recursive function to get a directory

        :param directory: the remote directory to get
        :param _root: locally, the directory to store any found files
        """
        size = 0
        if not _root.exists():
            _root.mkdir()
        # ls the directory
        for x in self.ls(parent=directory):
            if x[0] in [".", ".."]:
                # Discard . and ..
                continue
            remote = directory / x[0]
            local = _root / x[0]
            try:
                if x[1].FILE_ATTRIBUTE_DIRECTORY:
                    # Sub-directory
                    size += self._getr(remote, local)
                else:
                    # Sub-file
                    size += self.get(remote, local)[1]
                if _verb:
                    print(remote)
            except ValueError as ex:
                if _verb:
                    print(conf.color_theme.red(remote), "->", str(ex))
        return size

    @CLIUtil.addcommand(spaces=True, globsupport=True)
    def get(self, file, _dest=None, _verb=True, *, r=False):
        """
        Retrieve a file
        -r: recursively download a directory
        """
        if self._require_share():
            return
        if r:
            dirpar, dirname = self._parsepath(file)
            return file, self._getr(
                dirpar / dirname,  # Remotely
                _root=self.localpwd / dirname,  # Locally
                _verb=_verb,
            )
        else:
            fname = pathlib.PureWindowsPath(file).name
            # Write the buffer
            if _dest is None:
                _dest = self.localpwd / fname
            with _dest.open("wb") as fd:
                size = self._get_file(file, fd)
            return fname, size

    @CLIUtil.addoutput(get)
    def get_output(self, info):
        """
        Print the output of 'get'
        """
        print("Retrieved '%s' of size %s" % (info[0], human_size(info[1])))

    @CLIUtil.addcomplete(get)
    def get_complete(self, file):
        """
        Auto-complete get
        """
        if self._require_share(silent=True):
            return []
        return self._fs_complete(file)

    @CLIUtil.addcommand(spaces=True, globsupport=True)
    def cat(self, file):
        """
        Print a file
        """
        if self._require_share():
            return
        # Write the buffer to buffer
        buf = io.BytesIO()
        self._get_file(file, buf)
        return buf.getvalue()

    @CLIUtil.addoutput(cat)
    def cat_output(self, result):
        """
        Print the output of 'cat'
        """
        print(result.decode(errors="backslashreplace"))

    @CLIUtil.addcomplete(cat)
    def cat_complete(self, file):
        """
        Auto-complete cat
        """
        if self._require_share(silent=True):
            return []
        return self._fs_complete(file)

    @CLIUtil.addcommand(spaces=True)
    def put(self, file):
        """
        Upload a file
        """
        if self._require_share():
            return
        local_file = self.localpwd / file
        if local_file.is_dir():
            # Directory
            raise ValueError("put on dir not impl")
        else:
            fname = pathlib.Path(file).name
            with local_file.open("rb") as fd:
                size = self._send_file(fname, fd)
        self.ls_cache.clear()
        return fname, size

    @CLIUtil.addcomplete(put)
    def put_complete(self, folder):
        """
        Auto-complete put
        """
        return self._lfs_complete(folder, lambda x: not x.is_dir())

    @CLIUtil.addcommand(spaces=True)
    def rm(self, file):
        """
        Delete a file
        """
        if self._require_share():
            return
        # Get pwd of the ls
        fpath = self.pwd / file
        self.smbsock.set_TID(self.current_tree)
        # Open file
        fileId = self.smbsock.create_request(
            self.normalize_path(fpath),
            type="file",
            mode="d",
            extra_create_options=self.extra_create_options,
        )
        # Close the file
        self.smbsock.close_request(fileId)
        self.ls_cache.clear()
        return fpath.name

    @CLIUtil.addcomplete(rm)
    def rm_complete(self, file):
        """
        Auto-complete rm
        """
        if self._require_share(silent=True):
            return []
        return self._fs_complete(file)

    @CLIUtil.addcommand()
    def backup(self):
        """
        Turn on or off backup intent
        """
        if "FILE_OPEN_FOR_BACKUP_INTENT" in self.extra_create_options:
            print("Backup Intent: Off")
            self.extra_create_options.remove("FILE_OPEN_FOR_BACKUP_INTENT")
        else:
            print("Backup Intent: On")
            self.extra_create_options.append("FILE_OPEN_FOR_BACKUP_INTENT")

    @CLIUtil.addcommand(spaces=True)
    def watch(self, folder):
        """
        Watch file changes in folder (recursively)
        """
        if self._require_share():
            return
        # Get pwd of the ls
        fpath = self.pwd / folder
        self.smbsock.set_TID(self.current_tree)
        # Open file
        fileId = self.smbsock.create_request(
            self.normalize_path(fpath),
            type="folder",
            extra_create_options=self.extra_create_options,
        )
        print("Watching '%s'" % fpath)
        # Watch for changes
        try:
            while True:
                changes = self.smbsock.changenotify(fileId)
                for chg in changes:
                    print(chg.sprintf("%.time%: %Action% %FileName%"))
        except KeyboardInterrupt:
            pass
        # Close the file
        self.smbsock.close_request(fileId)
        print("Cancelled.")

    @CLIUtil.addcommand(spaces=True)
    def getsd(self, file):
        """
        Get the Security Descriptor
        """
        if self._require_share():
            return
        fpath = self.pwd / file
        self.smbsock.set_TID(self.current_tree)
        # Open file
        fileId = self.smbsock.create_request(
            self.normalize_path(fpath),
            type="",
            mode="",
            extra_desired_access=["READ_CONTROL", "ACCESS_SYSTEM_SECURITY"],
        )
        # Get the file size
        info = self.smbsock.query_info(
            FileId=fileId,
            InfoType="SMB2_0_INFO_SECURITY",
            FileInfoClass=0,
            AdditionalInformation=(
                0x00000001
                | 0x00000002
                | 0x00000004
                | 0x00000008
                | 0x00000010
                | 0x00000020
                | 0x00000040
                | 0x00010000
            ),
        )
        self.smbsock.close_request(fileId)
        return info

    @CLIUtil.addcomplete(getsd)
    def getsd_complete(self, file):
        """
        Auto-complete getsd
        """
        if self._require_share(silent=True):
            return []
        return self._fs_complete(file)

    @CLIUtil.addoutput(getsd)
    def getsd_output(self, results):
        """
        Print the output of 'getsd'
        """
        sd = SECURITY_DESCRIPTOR(results)
        print("Owner:", sd.OwnerSid.summary())
        print("Group:", sd.GroupSid.summary())
        if getattr(sd, "DACL", None):
            print("DACL:")
            for ace in sd.DACL.Aces:
                print(" - ", ace.toSDDL())
        if getattr(sd, "SACL", None):
            print("SACL:")
            for ace in sd.SACL.Aces:
                print(" - ", ace.toSDDL())


if __name__ == "__main__":
    from scapy.utils import AutoArgparse

    AutoArgparse(smbclient)
